# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import socket
from typing import Dict, List, Optional, Tuple, Union

from volatility3.framework import exceptions, interfaces, objects
from volatility3.framework.renderers import conversion

vollog = logging.getLogger(__name__)


def inet_ntop(address_family: int, packed_ip: Union[List[int], objects.Array]) -> str:
    if address_family in [socket.AF_INET6, socket.AF_INET]:
        try:
            return socket.inet_ntop(address_family, bytes(packed_ip))
        except AttributeError:
            raise RuntimeError(
                "This version of python does not have socket.inet_ntop, please upgrade"
            )
    raise OSError("[Errno 97] Address family not supported by protocol")


# Python's socket.AF_INET6 is 0x1e but Microsoft defines it
# as a constant value of 0x17 in their source code. Thus we
# need Microsoft's since that's what is found in memory.
AF_INET = 2
AF_INET6 = 0x17

# String representations of INADDR_ANY and INADDR6_ANY
inaddr_any = inet_ntop(socket.AF_INET, [0] * 4)
inaddr6_any = inet_ntop(socket.AF_INET6, [0] * 16)


class _TCP_LISTENER(objects.StructType):
    """Class for objects found in TcpL pools.

    This class serves as a base class for all pooled network objects.

    It exposes some functions which return sanity-checked members. Substructures referred to by a
    pointer may appear valid at first glance but will throw an InvalidAddressException on access.

    This is not a problem when objects are validated via their `is_valid()` method, but when
    scanning for semi-corrupted data this check will not be performed.

    Be mindful that most of those methods return `None` when they would access invalid data.
    If you want to process the raw data access the attributes directly, e.g.
    via `network_object.InetAF` instead of `network_object.get_address_family()`.

    """

    MIN_CREATETIME_YEAR = 1950
    MAX_CREATETIME_YEAR = 2200

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        type_name: str,
        object_info: interfaces.objects.ObjectInformation,
        size: int,
        members: Dict[str, Tuple[int, interfaces.objects.Template]],
    ) -> None:
        super().__init__(
            context=context,
            type_name=type_name,
            object_info=object_info,
            size=size,
            members=members,
        )

    def get_address_family(self):
        try:
            return self.InetAF.dereference().AddressFamily

        except exceptions.InvalidAddressException:
            return None

    def get_owner(self):
        try:
            return self.member("Owner").dereference()

        except exceptions.InvalidAddressException:
            return None

    def get_owner_pid(self) -> Optional[int]:
        owner = self.get_owner()

        if owner is None:
            return None

        if owner.is_valid():
            if owner.has_valid_member("UniqueProcessId"):
                return owner.UniqueProcessId

        return None

    def get_owner_procname(self) -> Optional[str]:
        owner = self.get_owner()

        if owner is None:
            return None

        if owner.is_valid():
            if owner.has_valid_member("ImageFileName"):
                return owner.ImageFileName.cast(
                    "string",
                    max_length=owner.ImageFileName.vol.count,
                    errors="replace",
                )

        return None

    def get_create_time(self):
        dt_obj = conversion.wintime_to_datetime(self.CreateTime.QuadPart)

        if isinstance(dt_obj, interfaces.renderers.BaseAbsentValue):
            return dt_obj

        # return None if the timestamp seems invalid
        if not (self.MIN_CREATETIME_YEAR < dt_obj.year < self.MAX_CREATETIME_YEAR):
            return None
        else:
            return dt_obj

    def get_in_addr(self):
        try:
            local_addr = self.LocalAddr.dereference()
            # there is a rare edge case here we have to consider:
            # if the struct has a null pointer at the LocalAddr offset,
            # this generally means this struct has no associated local address.
            # however, sometimes a pointer to the offset of 0 can be valid because
            # it points to a valid virtual memory address of 0. this confuses this
            # plugin because trying to access the nullpointer does not raise any
            # errors, leading to errors later down the line when accessing the
            # pointed-to _IN_ADDR addr4/6 attributes.

            # addr4/6 are at the same offset, accessing the first byte covers both.
            # if this causes no error, we can expect a valid network addr.
            _ = local_addr.pData.dereference().addr4[0]

            if local_addr.pData.dereference():
                inaddr = local_addr.inaddr
                return inaddr
            else:
                return None

        except exceptions.InvalidAddressException:
            return None

    def dual_stack_sockets(self):
        """Handle Windows dual-stack sockets"""

        # If this pointer is valid, the socket is bound to
        # a specific IP address. Otherwise, the socket is
        # listening on all IP addresses of the address family.

        # Note the remote address is always INADDR_ANY or
        # INADDR6_ANY for sockets. The moment a client
        # connects to the listener, a TCP_ENDPOINT is created
        # and that structure contains the remote address.

        inaddr = self.get_in_addr()

        if inaddr:
            if self.get_address_family() == AF_INET:
                yield "v4", inet_ntop(socket.AF_INET, inaddr.addr4), inaddr_any
            elif self.get_address_family() == AF_INET6:
                yield "v6", inet_ntop(socket.AF_INET6, inaddr.addr6), inaddr6_any
        else:
            yield "v4", inaddr_any, inaddr_any
            if self.get_address_family() == AF_INET6:
                yield "v6", inaddr6_any, inaddr6_any

    def is_valid(self):
        try:
            if self.get_address_family() not in (AF_INET, AF_INET6):
                vollog.debug(
                    f"netw obj 0x{self.vol.offset:x} invalid due to invalid address_family {self.get_address_family()}"
                )
                return False

        except exceptions.InvalidAddressException:
            vollog.debug(
                f"netw obj 0x{self.vol.offset:x} invalid due to invalid address access"
            )
            return False
        return True


class _TCP_ENDPOINT(_TCP_LISTENER):
    """Class for objects found in TcpE pools"""

    def _ipv4_or_ipv6(self, inaddr):
        if self.get_address_family() == AF_INET:
            return inet_ntop(socket.AF_INET, inaddr.addr4)
        else:
            return inet_ntop(socket.AF_INET6, inaddr.addr6)

    def get_local_address(self):
        try:
            inaddr = self.AddrInfo.dereference().Local.pData.dereference().dereference()

            return self._ipv4_or_ipv6(inaddr)

        except exceptions.InvalidAddressException:
            return None

    def get_remote_address(self):
        try:
            inaddr = self.AddrInfo.dereference().Remote.dereference()

            return self._ipv4_or_ipv6(inaddr)

        except exceptions.InvalidAddressException:
            return None

    def is_valid(self):
        # netstat calls this before validating the object itself
        try:
            state = self.State
        except exceptions.InvalidAddressException:
            return False

        if state not in state.choices.values():
            vollog.debug(
                f"{type(self)} 0x{self.vol.offset:x} invalid due to invalid tcp state {self.State}"
            )
            return False

        try:
            if self.get_address_family() not in (AF_INET, AF_INET6):
                vollog.debug(
                    f"{type(self)} 0x{self.vol.offset:x} invalid due to invalid address_family {self.get_address_family()}"
                )
                return False

            if not self.get_local_address() and (
                not self.get_owner()
                or self.get_owner().UniqueProcessId == 0
                or self.get_owner().UniqueProcessId > 65535
            ):
                vollog.debug(
                    f"{type(self)} 0x{self.vol.offset:x} invalid due to invalid owner data"
                )
                return False

        except exceptions.InvalidAddressException:
            vollog.debug(
                f"{type(self)} 0x{self.vol.offset:x} invalid due to invalid address access"
            )
            return False

        return True


class _UDP_ENDPOINT(_TCP_LISTENER):
    """Class for objects found in UdpA pools"""


class _LOCAL_ADDRESS(objects.StructType):
    @property
    def inaddr(self):
        return self.pData.dereference().dereference()


class _LOCAL_ADDRESS_WIN10_UDP(objects.StructType):
    @property
    def inaddr(self):
        return self.pData.dereference()


win10_x64_class_types = {
    "_TCP_ENDPOINT": _TCP_ENDPOINT,
    "_TCP_LISTENER": _TCP_LISTENER,
    "_UDP_ENDPOINT": _UDP_ENDPOINT,
    "_LOCAL_ADDRESS": _LOCAL_ADDRESS,
    "_LOCAL_ADDRESS_WIN10_UDP": _LOCAL_ADDRESS_WIN10_UDP,
}

class_types = {
    "_TCP_ENDPOINT": _TCP_ENDPOINT,
    "_TCP_LISTENER": _TCP_LISTENER,
    "_UDP_ENDPOINT": _UDP_ENDPOINT,
    "_LOCAL_ADDRESS": _LOCAL_ADDRESS,
}



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Generator, List, Union, Tuple
from volatility3.framework import objects, interfaces
from volatility3.framework import constants

vollog = logging.getLogger(__name__)


class ROW(objects.StructType):
    """A Row Structure."""

    def _valid_dbcs(self, dbcs_attr: int, text_attr_msb: int) -> bool:
        # TODO this need more research and testing
        # https://github.com/search?q=repo%3Amicrosoft%2Fterminal+DbcsAttr&type=code
        valid = text_attr_msb == 0 and dbcs_attr in (
            0x0,
            0x1,
            0x2,
            0x8,
            0x10,
            0x18,
            0x20,
            0x28,
            0x30,
            0x48,
            0x50,
            0x58,
            0x60,
            0x68,
            0x70,
            0x78,
            0x80,
            0x88,
            0xA8,
            0xB8,
            0xC0,
            0xC8,
            0x98,
            0xD8,
            0xE0,
            0xE8,
            0xF8,
            0xF0,
            0xA0,
        )
        if text_attr_msb == 0 and not valid:
            vollog.debug(f"Bad Dbcs Attribute {dbcs_attr:#x}")
        return valid

    def get_text(self, truncate: bool = True) -> str:
        """A convenience method to extract the text from the _ROW.  The _ROW
        contains a pointer CharRow to an array of CharRowCell objects. Each
        CharRowCell contains the wide character and an attribute. Enumerating
        self.CharRow.Chars and casting each character to unicode takes too long,
        so this reads the whole row into a buffer, then extracts the text characters."""

        layer = self._context.layers[self.vol.layer_name]
        offset = self.CharRow.Chars.vol.offset
        length = self.RowLength * 3
        char_row = layer.read(offset, length)
        line = ""
        try:
            if char_row:
                line = "".join(
                    (
                        char_row[i : i + 2].decode("utf-16le", errors="replace")
                        if self._valid_dbcs(char_row[i + 2], char_row[i + 1])
                        else ""
                    )
                    for i in range(0, len(char_row), 3)
                )
        except Exception:
            line = ""

        if truncate:
            return line.rstrip()
        else:
            return line


class ALIAS(objects.StructType):
    """An Alias Structure"""

    def get_source(self) -> Union[str, None]:
        return self.Source.get_command_string()

    def get_target(self) -> Union[str, None]:
        return self.Target.get_command_string()


class EXE_ALIAS_LIST(objects.StructType):
    """An Exe Alias List Structure"""

    def get_exename(self) -> Union[str, None]:
        exe_name = self.ExeName
        # Windows 10 22000 and Server 20348 removed the Pointer
        if isinstance(exe_name, objects.Pointer):
            exe_name = exe_name.dereference()
            return exe_name.get_string()

        return exe_name.get_command_string()

    def get_aliases(self) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """Generator for the individual aliases for a
        particular executable."""
        yield from self.AliasList.to_list(
            f"{self.get_symbol_table_name()}{constants.BANG}_ALIAS",
            "ListEntry",
        )


class SCREEN_INFORMATION(objects.StructType):
    """A Screen Information Structure."""

    @property
    def ScreenX(self) -> int:
        # 22000 change from an array of pointers to _ROW to an array of _ROW
        row = self.TextBufferInfo.BufferRows.Rows[0]
        if hasattr(row, "Row"):
            return row.Row.RowLength2
        else:
            return row.RowLength2

    @property
    def ScreenY(self) -> int:
        return self.TextBufferInfo.BufferCapacity

    def _truncate_rows(self, rows: List[str]) -> List[str]:
        """To truncate empty rows at the end, walk the list
        backwards and get the last non-empty row. Use that
        row index to splice. Rows are created based on the
        length given in the ROW structure, so empty rows will
        be ''."""

        non_empty_index = 0
        rows_traversed = False

        for index, row in enumerate(reversed(rows)):
            # the string was created based on the length in the ROW structure so it shouldn't have any bad data
            if len(row.rstrip()) > 0:
                non_empty_index = index
                break
            rows_traversed = True

        if non_empty_index == 0 and rows_traversed:
            rows = []
        else:
            rows = rows[0 : len(rows) - non_empty_index]

        return rows

    def get_buffer(
        self, truncate_rows: bool = True, truncate_lines: bool = True
    ) -> List[str]:
        """Get the screen buffer.

        The screen buffer is comprised of the screen's Y
        coordinate which tells us the number of rows and
        the X coordinate which tells us the width of each
        row in characters. Windows 10 17763 changed from
        a large text buffer to a grid of cells, with each
        cell containing a single wide character in that
        cell, stored in a CharRowCell object.

        @param truncate: True if the empty rows at the
        end (i.e. bottom) of the screen buffer should be
        suppressed.
        """
        rows = []

        capacity = self.TextBufferInfo.BufferCapacity
        start = self.TextBufferInfo.BufferStart
        buffer_rows = self.TextBufferInfo.BufferRows
        buffer_rows.Rows.count = self.TextBufferInfo.BufferCapacity

        for i in range(capacity):
            index = (start + i) % capacity
            row = buffer_rows.Rows[index]
            if hasattr(row, "Row"):
                row = row.Row
            try:
                text = row.get_text(truncate_lines)
                rows.append(text)
            except Exception:
                break

        if truncate_rows:
            rows = self._truncate_rows(rows)

        return rows


class CONSOLE_INFORMATION(objects.StructType):
    """A Console Information Structure."""

    @property
    def ScreenBuffer(self) -> interfaces.objects.ObjectInterface:
        return self.GetScreenBuffer

    def is_valid(self, max_buffers: int = 4) -> bool:
        """Determine if the structure is valid."""

        # Last displayed must be between -1 and max
        if self.HistoryBufferCount < 1 or self.HistoryBufferCount > max_buffers:
            return False

        if not self.get_title() and not self.get_original_title():
            return False

        return True

    def get_screens(self) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """Generator for screens in the console.

        A console can have multiple screen buffers at a time,
        but only the current/active one is displayed.

        Multiple screens are tracked using the singly-linked
        list _SCREEN_INFORMATION.Next.

        See CreateConsoleScreenBuffer
        """
        screens = [self.CurrentScreenBuffer]

        if self.ScreenBuffer not in screens:
            screens.append(self.ScreenBuffer)

        seen = set()

        for screen in screens:
            cur = screen
            while cur and cur.vol.offset != 0 and cur.vol.offset not in seen:
                cur.TextBufferInfo.BufferRows.Rows.count = (
                    cur.TextBufferInfo.BufferCapacity
                )
                yield cur
                seen.add(cur.vol.offset)
                cur = cur.Next

    def get_histories(
        self,
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        yield from self.HistoryList.to_list(
            f"{self.get_symbol_table_name()}{constants.BANG}_COMMAND_HISTORY",
            "ListEntry",
        )

    def get_exe_aliases(
        self,
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        exe_alias_list = self.ExeAliasList
        # Windows 10 22000 and Server 20348 made this a Pointer
        if isinstance(exe_alias_list, objects.Pointer):
            exe_alias_list = exe_alias_list.dereference()
        yield from exe_alias_list.to_list(
            f"{self.get_symbol_table_name()}{constants.BANG}_EXE_ALIAS_LIST",
            "ListEntry",
        )

    def get_processes(
        self,
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        yield from self.ConsoleProcessList.to_list(
            f"{self.get_symbol_table_name()}{constants.BANG}_CONSOLE_PROCESS_LIST",
            "ListEntry",
        )

    def get_title(self) -> Union[str, None]:
        try:
            return self.Title.dereference().cast(
                "string", encoding="utf-16", errors="replace", max_length=512
            )
        except Exception:
            return ""

    def get_original_title(self) -> Union[str, None]:
        try:
            return self.OriginalTitle.dereference().cast(
                "string", encoding="utf-16", errors="replace", max_length=512
            )
        except Exception:
            return ""


class COMMAND(objects.StructType):
    """A Command Structure"""

    def is_valid(self) -> bool:
        if (
            self.Length < 1
            or self.Allocated < 1
            or self.Length > 1024
            or self.Allocated > 1024
        ):
            return False

        return True

    def get_command_string(self) -> Union[str, None]:
        if self.Length < 8:
            return self.Chars.cast(
                "string",
                encoding="utf-16",
                errors="replace",
                max_length=self.Length * 2,
            )
        elif self.Length < 1024:
            return self.Pointer.dereference().cast(
                "string",
                encoding="utf-16",
                errors="replace",
                max_length=self.Length * 2,
            )

        return None


class COMMAND_HISTORY(objects.StructType):
    """A Command History Structure."""

    @property
    def CommandCount(self) -> int:
        command_type = self.get_symbol_table_name() + constants.BANG + "_COMMAND"
        command_size = self._context.symbol_space.get_type(command_type).size
        return int((self.CommandBucket.End - self.CommandBucket.Begin) / command_size)

    @property
    def ProcessHandle(self) -> int:
        """Allow ProcessHandle to be referenced regardless of OS version"""
        return self.ConsoleProcessHandle.ProcessHandle

    def is_valid(self, max_history: int = 50) -> bool:
        # The count must be between zero and max
        if self.CommandCount < 0 or self.CommandCount > max_history:
            return False

        # Last displayed must be between -1 and max
        if self.LastDisplayed < -1 or self.LastDisplayed > max_history:
            return False

        # Process handle must be a valid pid
        if (
            self.ProcessHandle <= 0
            or self.ProcessHandle > 0xFFFF
            or self.ProcessHandle % 4 != 0
        ):
            return False

        return True

    def get_application(self) -> Union[str, None]:
        return self.Application.get_command_string()

    def scan_command_bucket(
        self, end: Union[int, None] = None
    ) -> Generator[Tuple[int, interfaces.objects.ObjectInterface], None, None]:
        """Brute force print all strings pointed to by the CommandBucket entries by
        going to greater of EndCapacity or CommandCountMax*sizeof(_COMMAND)"""

        command_type = self.get_symbol_table_name() + constants.BANG + "_COMMAND"
        command_history_size = self._context.symbol_space.get_type(
            self.vol.type_name
        ).size
        command_size = self._context.symbol_space.get_type(command_type).size

        if end is None:
            end = max(
                self.CommandBucket.EndCapacity,
                self.CommandBucket.Begin + command_history_size * self.CommandCountMax,
            )

        for i, pointer in enumerate(range(self.CommandBucket.Begin, end, command_size)):
            cmd = self._context.object(command_type, self.vol.layer_name, pointer)
            if cmd.is_valid():
                yield i, cmd

    def get_commands(
        self,
    ) -> Generator[Tuple[int, interfaces.objects.ObjectInterface], None, None]:
        """Generator for commands in the history buffer.

        The CommandBucket is an array of pointers to _COMMAND
        structures. The array size is CommandCount. Once CommandCount
        is reached, the oldest commands are cycled out and the
        rest are coalesced.
        """

        yield from self.scan_command_bucket(self.CommandBucket.End)


win10_x64_class_types = {
    "_EXE_ALIAS_LIST": EXE_ALIAS_LIST,
    "_ALIAS": ALIAS,
    "_ROW": ROW,
    "_SCREEN_INFORMATION": SCREEN_INFORMATION,
    "_CONSOLE_INFORMATION": CONSOLE_INFORMATION,
    "_COMMAND_HISTORY": COMMAND_HISTORY,
    "_COMMAND": COMMAND,
}
class_types = {
    "_ROW": ROW,
    "_SCREEN_INFORMATION": SCREEN_INFORMATION,
    "_CONSOLE_INFORMATION": CONSOLE_INFORMATION,
    "_COMMAND_HISTORY": COMMAND_HISTORY,
    "_COMMAND": COMMAND,
}



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import re
from typing import Dict, Generator, List, Set, Tuple, Optional

from volatility3.framework import interfaces, renderers, exceptions, constants
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import intel, resources, linear
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist

vollog = logging.getLogger(__name__)


class Strings(interfaces.plugins.PluginInterface):
    """Reads output from the strings command and indicates which process(es) each string belongs to."""

    _required_framework_version = (2, 0, 0)

    # 2.0.0 - change signature of `generate_mapping`
    _version = (2, 0, 0)

    strings_pattern = re.compile(rb"^(?:\W*)([0-9]+)(?:\W*)(\w[\w\W]+)\n?")

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process ID to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.URIRequirement(
                name="strings_file", description="Strings file"
            ),
        ]
        # TODO: Make URLRequirement that can accept a file address which the framework can open

    def run(self):
        return renderers.TreeGrid(
            [("String", str), ("Physical Address", format_hints.Hex), ("Result", str)],
            self._generator(),
        )

    def _generator(self) -> Generator[Tuple, None, None]:
        """Generates results from a strings file."""
        string_list: List[Tuple[int, bytes]] = []

        # Test strings file format is accurate
        accessor = resources.ResourceAccessor()
        strings_fp = accessor.open(self.config["strings_file"], "rb")
        line = strings_fp.readline()
        count: float = 0
        while line:
            count += 1
            try:
                offset, string = self._parse_line(line)
                string_list.append((offset, string))
            except ValueError:
                vollog.error(f"Line in unrecognized format: line {count}")
            line = strings_fp.readline()

        revmap = self.generate_mapping(
            context=self.context,
            kernel_module_name=self.config["kernel"],
            progress_callback=self._progress_callback,
            pid_list=self.config["pid"],
        )

        last_prog: float = 0
        line_count: float = 0
        num_strings = len(string_list)
        for offset, string in string_list:
            line_count += 1
            try:
                revmap_list = [
                    name + ":" + hex(offset) for (name, offset) in revmap[offset >> 12]
                ]
            except (IndexError, KeyError):
                revmap_list = ["FREE MEMORY"]
            yield (
                0,
                (
                    str(string, "latin-1"),
                    format_hints.Hex(offset),
                    ", ".join(revmap_list),
                ),
            )
            prog = line_count / num_strings * 100
            if round(prog, 1) > last_prog:
                last_prog = round(prog, 1)
                self._progress_callback(prog, "Matching strings in memory")

    def _parse_line(self, line: bytes) -> Tuple[int, bytes]:
        """Parses a single line from a strings file.

        Args:
            line: bytes of the line of a strings file (an offset and a string)

        Returns:
            Tuple of the offset and the string found at that offset
        """

        match = self.strings_pattern.search(line)
        if not match:
            raise ValueError("Strings file contains invalid strings line")
        offset, string = match.group(1, 2)
        return int(offset), string

    @classmethod
    def generate_mapping(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        progress_callback: constants.ProgressCallback = None,
        pid_list: Optional[List[int]] = None,
    ) -> Dict[int, Set[Tuple[str, int]]]:
        """Creates a reverse mapping between virtual addresses and physical
        addresses.

        Args:
            context: the context for the method to run against
            kernel_module_name: the name of the module forthe kernel
            progress_callback: an optional callable to display progress
            pid_list: a lit of process IDs to consider when generating the reverse map

        Returns:
            A mapping of virtual offsets to strings and physical offsets
        """
        filter = pslist.PsList.create_pid_filter(pid_list)

        kernel = context.modules[kernel_module_name]

        layer = context.layers[kernel.layer_name]
        reverse_map: Dict[int, Set[Tuple[str, int]]] = dict()
        if isinstance(layer, intel.Intel):
            # We don't care about errors, we just wanted chunks that map correctly
            for mapval in layer.mapping(0x0, layer.maximum_address, ignore_errors=True):
                offset, _, mapped_offset, mapped_size, maplayer = mapval
                for val in range(mapped_offset, mapped_offset + mapped_size, 0x1000):
                    cur_set = reverse_map.get(val >> 12, set())
                    cur_set.add(("kernel", offset))
                    reverse_map[val >> 12] = cur_set
                if progress_callback:
                    progress_callback(
                        (offset * 100) / layer.maximum_address,
                        "Creating reverse kernel map",
                    )

            # TODO: Include kernel modules

            for process in pslist.PsList.list_processes(
                context=context, kernel_module_name=kernel_module_name
            ):
                if not filter(process):
                    proc_id = "Unknown"
                    try:
                        proc_id = process.UniqueProcessId
                        proc_layer_name = process.add_process_layer()
                    except exceptions.InvalidAddressException as excp:
                        vollog.debug(
                            f"Process {proc_id}: invalid address {excp.invalid_address} in layer {excp.layer_name}"
                        )
                        continue

                    proc_layer = context.layers[proc_layer_name]
                    if isinstance(proc_layer, linear.LinearlyMappedLayer):
                        for mapval in proc_layer.mapping(
                            0x0, proc_layer.maximum_address, ignore_errors=True
                        ):
                            mapped_offset, _, offset, mapped_size, _maplayer = mapval
                            for val in range(
                                mapped_offset, mapped_offset + mapped_size, 0x1000
                            ):
                                cur_set = reverse_map.get(mapped_offset >> 12, set())
                                cur_set.add(
                                    (f"Process {process.UniqueProcessId}", offset)
                                )
                                reverse_map[mapped_offset >> 12] = cur_set
                            # FIXME: make the progress for all processes, rather than per-process
                            if progress_callback:
                                progress_callback(
                                    (offset * 100) / layer.maximum_address,
                                    f"Creating mapping for task {process.UniqueProcessId}",
                                )

        return reverse_map



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import abc
import datetime
import enum
import io
import json
import logging
import traceback
from typing import Generator, Iterable, List, Optional, Tuple, Type

from volatility3 import framework
from volatility3.framework import automagic, exceptions, interfaces, plugins, renderers
from volatility3.framework.configuration import requirements

vollog = logging.getLogger(__name__)


class TimeLinerType(enum.IntEnum):
    CREATED = 1
    MODIFIED = 2
    ACCESSED = 3
    CHANGED = 4


class TimeLinerInterface(
    interfaces.configuration.VersionableInterface, metaclass=abc.ABCMeta
):
    """Interface defining methods that timeliner will use to generate a body
    file."""

    _version = (1, 0, 0)

    @abc.abstractmethod
    def generate_timeline(
        self,
    ) -> Generator[Tuple[str, TimeLinerType, datetime.datetime], None, None]:
        """Method generates Tuples of (description, timestamp_type, timestamp)

        These need not be generated in any particular order, sorting
        will be done later
        """


class Timeliner(interfaces.plugins.PluginInterface):
    """Runs all relevant plugins that provide time related information and \
orders the results by time."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 1, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.timeline = {}
        self.usable_plugins = None
        self.automagics: Optional[List[interfaces.automagic.AutomagicInterface]] = None

    @classmethod
    def get_usable_plugins(
        cls, selected_list: Optional[List[str]] = None
    ) -> List[Type]:
        # Initialize for the run
        plugin_list = list(framework.class_subclasses(TimeLinerInterface))

        # Get the filter from the configuration
        def passthrough(name: str, selected: List[str]) -> bool:
            return True

        filter_func = passthrough
        if selected_list:

            def filter_plugins(name: str, selected: List[str]) -> bool:
                return any(s in name for s in selected)

            filter_func = filter_plugins
        else:
            selected_list = []

        return [
            plugin_class
            for plugin_class in plugin_list
            if filter_func(plugin_class.__name__, selected_list)
        ]

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.BooleanRequirement(
                name="record-config",
                description="Whether to record the state of all the plugins once complete",
                optional=True,
                default=False,
            ),
            requirements.ListRequirement(
                name="plugin-filter",
                description="Only run plugins featuring this substring",
                element_type=str,
                optional=True,
                default=[],
            ),
            requirements.BooleanRequirement(
                name="create-bodyfile",
                description="Whether to create a body file whilst producing results",
                optional=True,
                default=False,
            ),
        ]

    def _sort_function(self, item):
        data = item[1]

        def sortable(timestamp):
            max_date = datetime.datetime(
                day=1, month=12, year=datetime.MAXYEAR, tzinfo=datetime.timezone.utc
            )
            if isinstance(timestamp, interfaces.renderers.BaseAbsentValue):
                return max_date
            return timestamp

        return [sortable(timestamp) for timestamp in data[2:]]

    def _generator(
        self, runnable_plugins: List[TimeLinerInterface]
    ) -> Optional[Iterable[Tuple[int, Tuple]]]:
        """Takes a timeline, sorts it and output the data from each relevant
        row from each plugin."""
        # Generate the results for each plugin
        data = []

        # Open the bodyfile now, so we can start outputting to it immediately
        if self.config.get("create-bodyfile", True):
            file_data = self.open("volatility.body")
            fp = io.TextIOWrapper(file_data, write_through=True)
        else:
            file_data = None
            fp = None

        for plugin in runnable_plugins:
            plugin_name = plugin.__class__.__name__
            self._progress_callback(
                (runnable_plugins.index(plugin) * 100) // len(runnable_plugins),
                f"Running plugin {plugin_name}...",
            )
            try:
                vollog.log(logging.INFO, f"Running {plugin_name}")
                for item, timestamp_type, timestamp in plugin.generate_timeline():
                    times = self.timeline.get((plugin_name, item), {})
                    if times.get(timestamp_type, None) is not None:
                        vollog.debug(
                            f"Multiple timestamps for the same plugin/file combination found: {plugin_name} {item}"
                        )
                    times[timestamp_type] = timestamp
                    self.timeline[(plugin_name, item)] = times

                for plugin_name, item in self.timeline:
                    data.append(
                        (
                            0,
                            [
                                plugin_name,
                                item,
                                times.get(
                                    TimeLinerType.CREATED,
                                    renderers.NotApplicableValue(),
                                ),
                                times.get(
                                    TimeLinerType.MODIFIED,
                                    renderers.NotApplicableValue(),
                                ),
                                times.get(
                                    TimeLinerType.ACCESSED,
                                    renderers.NotApplicableValue(),
                                ),
                                times.get(
                                    TimeLinerType.CHANGED,
                                    renderers.NotApplicableValue(),
                                ),
                            ],
                        )
                    )

                    # Write each entry because the body file doesn't need to be sorted
                    if fp:
                        times = self.timeline[(plugin_name, item)]
                        # Body format is: MD5|name|inode|mode_as_string|UID|GID|size|atime|mtime|ctime|crtime

                        if self._any_time_present(times):
                            fp.write(
                                "|{} - {}|0|0|0|0|0|{}|{}|{}|{}\n".format(
                                    plugin_name,
                                    self._sanitize_body_format(item),
                                    self._text_format(
                                        times.get(TimeLinerType.ACCESSED, "0")
                                    ),
                                    self._text_format(
                                        times.get(TimeLinerType.MODIFIED, "0")
                                    ),
                                    self._text_format(
                                        times.get(TimeLinerType.CHANGED, "0")
                                    ),
                                    self._text_format(
                                        times.get(TimeLinerType.CREATED, "0")
                                    ),
                                )
                            )
            except Exception as e:
                vollog.log(
                    logging.INFO,
                    f"Exception occurred running plugin: {plugin_name}: {e}",
                )
                vollog.log(logging.DEBUG, traceback.format_exc())

        yield from sorted(data, key=self._sort_function)

        # Write out a body file if necessary
        if self.config.get("create-bodyfile", True):
            if fp:
                fp.close()
                file_data.close()

    def _sanitize_body_format(self, value):
        return value.replace("|", "_")

    def _any_time_present(self, times):
        for time in TimeLinerType:
            if not isinstance(
                times.get(time, renderers.NotApplicableValue),
                interfaces.renderers.BaseAbsentValue,
            ):
                return True
        return False

    def _text_format(self, value):
        """Formats a value as text, in case it is an AbsentValue"""
        if isinstance(value, interfaces.renderers.BaseAbsentValue):
            return "0"
        if isinstance(value, datetime.datetime):
            return int(value.timestamp())
        return value

    def run(self):
        """Isolate each plugin and run it."""

        # Use all the plugins if there's no filter
        self.usable_plugins = self.usable_plugins or self.get_usable_plugins()
        self.automagics = self.automagics or automagic.available(self._context)
        plugins_to_run = []
        requirement_configs = {}

        filter_list = self.config["plugin-filter"]
        # Identify plugins that we can run which output datetimes
        for plugin_class in self.usable_plugins:
            if not issubclass(plugin_class, TimeLinerInterface):
                # get_usable_plugins() should filter this, but adding a safeguard just in case
                continue

            if filter_list and not any(
                filter in plugin_class.__module__ + "." + plugin_class.__name__
                for filter in filter_list
            ):
                continue

            try:
                automagics = automagic.choose_automagic(self.automagics, plugin_class)

                for requirement in plugin_class.get_requirements():
                    if requirement.name in requirement_configs:
                        config_req, config_value = requirement_configs[requirement.name]
                        if requirement == config_req:
                            self.context.config[
                                interfaces.configuration.path_join(
                                    self.config_path, plugin_class.__name__
                                )
                            ] = config_value

                plugin = plugins.construct_plugin(
                    self.context,
                    automagics,
                    plugin_class,
                    self.config_path,
                    self._progress_callback,
                    self.open,
                )

                for requirement in plugin.get_requirements():
                    if requirement.name not in requirement_configs:
                        config_value = plugin.config.get(requirement.name, None)
                        if config_value:
                            requirement_configs[requirement.name] = (
                                requirement,
                                config_value,
                            )

                plugins_to_run.append(plugin)

            except exceptions.UnsatisfiedException as excp:
                # Remove the failed plugin from the list and continue
                vollog.debug(
                    f"Unable to satisfy {plugin_class.__name__}: {excp.unsatisfied}"
                )
                continue

        if self.config.get("record-config", False):
            total_config = {}
            for plugin in plugins_to_run:
                old_dict = dict(plugin.build_configuration())
                for entry in old_dict:
                    total_config[
                        interfaces.configuration.path_join(
                            plugin.__class__.__name__, entry
                        )
                    ] = old_dict[entry]

            with self.open("config.json") as file_data:
                with io.TextIOWrapper(file_data, write_through=True) as fp:
                    json.dump(total_config, fp, sort_keys=True, indent=2)

        return renderers.TreeGrid(
            columns=[
                ("Plugin", str),
                ("Description", str),
                ("Created Date", datetime.datetime),
                ("Modified Date", datetime.datetime),
                ("Accessed Date", datetime.datetime),
                ("Changed Date", datetime.datetime),
            ],
            generator=self._generator(plugins_to_run),
        )

    def build_configuration(self):
        """Builds the configuration to save for the plugin such that it can be
        reconstructed."""
        vollog.warning("Unable to record configuration data for the timeliner plugin")
        return []



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from volatility3.framework import interfaces, deprecation
from volatility3.plugins.windows.malware import skeleton_key_check

vollog = logging.getLogger(__name__)


class Skeleton_Key_Check(
    interfaces.plugins.PluginInterface,
    deprecation.PluginRenameClass,
    replacement_class=skeleton_key_check.Skeleton_Key_Check,
    removal_date="2026-06-07",
):
    """Looks for signs of Skeleton Key malware"""

    _required_framework_version = (2, 4, 0)
    _version = (1, 0, 0)



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import List, Optional, Type

from volatility3.framework import renderers, interfaces, constants, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins

vollog = logging.getLogger(__name__)


class LayerWriter(plugins.PluginInterface):
    """Runs the automagics and writes out the primary layer produced by the stacker."""

    default_block_size = 0x500000

    _required_framework_version = (2, 0, 0)
    _version = (2, 0, 1)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.TranslationLayerRequirement(
                name="primary", description="Memory layer for the kernel"
            ),
            requirements.IntRequirement(
                name="block_size",
                description="Size of blocks to copy over",
                default=cls.default_block_size,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="list",
                description="List available layers",
                default=False,
                optional=True,
            ),
            requirements.ListRequirement(
                name="layers",
                element_type=str,
                description="Names of layers to write (defaults to the highest non-mapped layer)",
                default=None,
                optional=True,
            ),
        ]

    @classmethod
    def write_layer(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        preferred_name: str,
        open_method: Type[plugins.FileHandlerInterface],
        chunk_size: Optional[int] = None,
        progress_callback: Optional[constants.ProgressCallback] = None,
    ) -> Optional[plugins.FileHandlerInterface]:
        """Produces a FileHandler from the named layer in the provided context or None on failure

        Args:
            context: the context from which to read the memory layer
            layer_name: the name of the layer to write out
            preferred_name: a string with the preferred filename for the file
            chunk_size: an optional size for the chunks that should be written (defaults to 0x500000)
            open_method: class for creating FileHandler context managers
            progress_callback: an optional function that takes a percentage and a string that displays output
        """

        if layer_name not in context.layers:
            raise exceptions.LayerException("Layer not found")
        layer = context.layers[layer_name]

        if chunk_size is None:
            chunk_size = cls.default_block_size

        file_handle = open_method(preferred_name)
        for i in range(0, layer.maximum_address, chunk_size):
            current_chunk_size = min(chunk_size, layer.maximum_address + 1 - i)
            data = layer.read(i, current_chunk_size, pad=True)
            file_handle.write(data)
            if progress_callback:
                progress_callback(
                    (i / layer.maximum_address) * 100, f"Writing layer {layer_name}"
                )
        return file_handle

    def _generator(self):
        if self.config["list"]:
            for name in self.context.layers:
                yield 0, (name,)
        else:
            # Choose the most recently added layer that isn't virtual
            if not self.config["layers"]:
                self.config["layers"] = []
                for name in self.context.layers:
                    if "mapped" not in self.context.layers[name].metadata:
                        self.config["layers"] = [name]

            for name in self.config["layers"]:
                # Check the layer exists and validate the output file
                if name not in self.context.layers:
                    yield 0, (f"Layer Name {name} does not exist",)
                else:
                    default_output_name = f"{name}.raw"
                    output_name = self.config.get("output", default_output_name)
                    try:
                        file_handle = self.write_layer(
                            self.context,
                            name,
                            output_name,
                            self.open,
                            self.config.get("block_size", self.default_block_size),
                            progress_callback=self._progress_callback,
                        )
                        file_handle.close()

                        # Update the filename, which may have changed if a file
                        # with the same name already existed.
                        output_name = file_handle.preferred_filename
                    except OSError as excp:
                        yield 0, (f"Layer cannot be written to {output_name}: {excp}",)

                    yield 0, (f"Layer has been written to {output_name}",)

    def _generate_layers(self):
        """List layer names from this run"""
        for name in self.context.layers:
            yield (0, (name, self.context.layers[name].__class__.__name__))

    def run(self):
        if self.config["list"]:
            return renderers.TreeGrid(
                [("Layer name", str), ("Layer type", str)], self._generate_layers()
            )
        return renderers.TreeGrid([("Status", str)], self._generator())



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import datetime
import logging
from typing import Optional, Tuple, Union, List, Dict
from volatility3.framework import constants, interfaces

vollog = logging.getLogger(__name__)


class ProducerMetadata(interfaces.symbols.MetadataInterface):
    """Class to handle the Producer metadata from an ISF"""

    @property
    def name(self) -> Optional[str]:
        return self._json_data.get("name", None)

    @property
    def version_string(self) -> str:
        """Returns the ISF file producer's version as a string.
        If no version is present, an empty string is returned.
        """
        return self._json_data.get("version", "")

    @property
    def version(self) -> Optional[Tuple[int, ...]]:
        """Returns the version of the ISF file producer"""
        version = self.version_string
        if not version:
            return None
        if all(x in "0123456789." for x in version):
            return tuple(int(x) for x in version.split("."))
        vollog.log(
            constants.LOGLEVEL_VVVV,
            f"Metadata version contains unexpected characters: '{version}'",
        )
        return None

    @property
    def datetime(self) -> Optional[datetime.datetime]:
        """Returns a timestamp for when the file was produced"""
        if "datetime" not in self._json_data:
            return None
        try:
            timestamp = datetime.datetime.strptime(
                self._json_data["datetime"], "YYYY-MM-DD"
            )
        except (TypeError, ValueError):
            vollog.debug("Invalid timestamp in producer information of symbol table")
            return None
        return timestamp


class WindowsMetadata(interfaces.symbols.MetadataInterface):
    """Class to handle the metadata from a Windows symbol table."""

    @property
    def pe_version(
        self,
    ) -> Optional[Union[Tuple[int, int, int], Tuple[int, int, int, int]]]:
        build = self._json_data.get("pe", {}).get("build", None)
        revision = self._json_data.get("pe", {}).get("revision", None)
        minor = self._json_data.get("pe", {}).get("minor", None)
        major = self._json_data.get("pe", {}).get("major", None)
        if revision is None or minor is None or major is None:
            return None
        if build is None:
            return major, minor, revision
        return major, minor, revision, build

    @property
    def pe_version_string(self) -> Optional[str]:
        if self.pe_version is None:
            return None
        return ".".join(self.pe_version)

    @property
    def pdb_guid(self) -> Optional[str]:
        return self._json_data.get("pdb", {}).get("GUID", None)

    @property
    def pdb_age(self) -> Optional[int]:
        return self._json_data.get("pdb", {}).get("age", None)


class PosixMetadata(interfaces.symbols.MetadataInterface):
    """Base class to handle metadata of Posix-based ISF sources"""

    def get_types_sources(self) -> List[Optional[Dict]]:
        """Returns the types sources metadata"""
        return self._json_data.get("types", [])

    def get_symbols_sources(self) -> List[Optional[Dict]]:
        """Returns the symbols sources metadata"""
        return self._json_data.get("symbols", [])


class LinuxMetadata(PosixMetadata):
    """Class to handle the metadata from a Linux symbol table."""


class MacMetadata(PosixMetadata):
    """Class to handle the metadata from a Mac symbol table."""



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging

from volatility3.framework import constants
from volatility3.framework import renderers, exceptions, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import ssdt, driverscan, modules

vollog = logging.getLogger(__name__)

MAJOR_FUNCTIONS = [
    "IRP_MJ_CREATE",
    "IRP_MJ_CREATE_NAMED_PIPE",
    "IRP_MJ_CLOSE",
    "IRP_MJ_READ",
    "IRP_MJ_WRITE",
    "IRP_MJ_QUERY_INFORMATION",
    "IRP_MJ_SET_INFORMATION",
    "IRP_MJ_QUERY_EA",
    "IRP_MJ_SET_EA",
    "IRP_MJ_FLUSH_BUFFERS",
    "IRP_MJ_QUERY_VOLUME_INFORMATION",
    "IRP_MJ_SET_VOLUME_INFORMATION",
    "IRP_MJ_DIRECTORY_CONTROL",
    "IRP_MJ_FILE_SYSTEM_CONTROL",
    "IRP_MJ_DEVICE_CONTROL",
    "IRP_MJ_INTERNAL_DEVICE_CONTROL",
    "IRP_MJ_SHUTDOWN",
    "IRP_MJ_LOCK_CONTROL",
    "IRP_MJ_CLEANUP",
    "IRP_MJ_CREATE_MAILSLOT",
    "IRP_MJ_QUERY_SECURITY",
    "IRP_MJ_SET_SECURITY",
    "IRP_MJ_POWER",
    "IRP_MJ_SYSTEM_CONTROL",
    "IRP_MJ_DEVICE_CHANGE",
    "IRP_MJ_QUERY_QUOTA",
    "IRP_MJ_SET_QUOTA",
    "IRP_MJ_PNP",
]


class DriverIrp(interfaces.plugins.PluginInterface):
    """List IRPs for drivers in a particular windows memory image."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="ssdt", component=ssdt.SSDT, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="driverscan", component=driverscan.DriverScan, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="modules", component=modules.Modules, version=(3, 0, 0)
            ),
        ]

    def _generator(self):
        collection = ssdt.SSDT.build_module_collection(
            context=self.context,
            kernel_module_name=self.config["kernel"],
        )

        kernel_space_start = modules.Modules.get_kernel_space_start(
            context=self.context,
            module_name=self.config["kernel"],
        )

        for driver in driverscan.DriverScan.scan_drivers(
            self.context,
            self.config["kernel"],
        ):
            try:
                driver_name = driver.get_driver_name()
            except (ValueError, exceptions.InvalidAddressException):
                driver_name = renderers.NotApplicableValue()

            for i in range(len(driver.MajorFunction)):
                try:
                    irp_handler = driver.MajorFunction[i]
                except exceptions.InvalidAddressException:
                    vollog.debug(
                        f"Failed to get IRP handler entry at index {i} for driver at {driver.vol.offset:#x}"
                    )
                    continue

                # smear
                if irp_handler < kernel_space_start:
                    continue

                module_symbols = collection.get_module_symbols_by_absolute_location(
                    irp_handler
                )

                module_found = False

                for module_name, symbol_generator in module_symbols:
                    module_found = True
                    symbols_found = False

                    for symbol in symbol_generator:
                        symbols_found = True
                        yield (
                            0,
                            (
                                format_hints.Hex(driver.vol.offset),
                                driver_name,
                                MAJOR_FUNCTIONS[i],
                                format_hints.Hex(irp_handler),
                                module_name,
                                symbol.split(constants.BANG)[1],
                            ),
                        )

                    if not symbols_found:
                        yield (
                            0,
                            (
                                format_hints.Hex(driver.vol.offset),
                                driver_name,
                                MAJOR_FUNCTIONS[i],
                                format_hints.Hex(irp_handler),
                                module_name,
                                renderers.NotAvailableValue(),
                            ),
                        )

                if not module_found:
                    yield (
                        0,
                        (
                            format_hints.Hex(driver.vol.offset),
                            driver_name,
                            MAJOR_FUNCTIONS[i],
                            format_hints.Hex(irp_handler),
                            renderers.NotAvailableValue(),
                            renderers.NotAvailableValue(),
                        ),
                    )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Driver Name", str),
                ("IRP", str),
                ("Address", format_hints.Hex),
                ("Module", str),
                ("Symbol", str),
            ],
            self._generator(),
        )



# This file is Copyright 2021 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from volatility3.framework import interfaces, constants
from volatility3.framework import objects


class SUMMARY_DUMP(objects.StructType):
    def get_buffer(
        self, sub_type: str, count: int
    ) -> interfaces.objects.ObjectInterface:
        symbol_table_name = self.get_symbol_table_name()
        subtype = self._context.symbol_space.get_type(
            symbol_table_name + constants.BANG + sub_type
        )
        return self._context.object(
            object_type=symbol_table_name + constants.BANG + "array",
            layer_name=self.vol.layer_name,
            offset=self.BufferChar.vol.offset,
            count=count,
            subtype=subtype,
        )

    def get_buffer_char(self) -> interfaces.objects.ObjectInterface:
        return self.get_buffer(
            sub_type="unsigned char", count=(self.BitmapSize + 7) // 8
        )

    def get_buffer_long(self) -> interfaces.objects.ObjectInterface:
        return self.get_buffer(
            sub_type="unsigned long", count=(self.BitmapSize + 31) // 32
        )


class_types = {"_SUMMARY_DUMP": SUMMARY_DUMP}



# This file is Copyright 2023 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from typing import List
from volatility3.framework import interfaces, renderers, constants
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.symbols.linux import network
from volatility3.framework.symbols.linux.extensions import network as net_extensions


class Addr(plugins.PluginInterface):
    """Lists network interface information for all devices"""

    _required_framework_version = (2, 22, 0)

    _version = (1, 0, 2)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="Net", component=network.NetSymbols, version=(1, 0, 0)
            ),
        ]

    def _gather_net_dev_info(self, net_dev: net_extensions.net_device):
        mac_addr = net_dev.get_mac_address()
        promisc = net_dev.promisc
        operational_state = net_dev.get_operational_state()
        iface_name = net_dev.get_device_name()
        iface_ifindex = net_dev.ifindex
        try:
            net_ns_id = net_dev.get_net_namespace_id()
        except AttributeError:
            net_ns_id = None

        # Interface IPv4 Addresses
        in_device = net_dev.ip_ptr.dereference().cast("in_device")
        for in_ifaddr in in_device.get_addresses():
            prefix_len = in_ifaddr.get_prefix_len()
            scope_type = in_ifaddr.get_scope_type()
            ip_addr = in_ifaddr.get_address()
            yield (
                net_ns_id,
                iface_ifindex,
                iface_name,
                mac_addr,
                promisc,
                ip_addr,
                prefix_len,
                scope_type,
                operational_state,
            )

        # Interface IPv6 Addresses
        inet6_dev = net_dev.ip6_ptr.dereference().cast("inet6_dev")
        for inet6_ifaddr in inet6_dev.get_addresses():
            prefix_len = inet6_ifaddr.get_prefix_len()
            scope_type = inet6_ifaddr.get_scope_type()
            ip6_addr = inet6_ifaddr.get_address()
            yield (
                net_ns_id,
                iface_ifindex,
                iface_name,
                mac_addr,
                promisc,
                ip6_addr,
                prefix_len,
                scope_type,
                operational_state,
            )

    def _enumerate_net_namespace_list(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        net_type_symname = vmlinux.symbol_table_name + constants.BANG + "net"
        net_device_symname = vmlinux.symbol_table_name + constants.BANG + "net_device"
        network.NetSymbols.apply(self.context.symbol_space[vmlinux.symbol_table_name])

        # 'net_namespace_list' exists from kernels >= 2.6.24
        net_namespace_list = vmlinux.object_from_symbol("net_namespace_list")
        for net_ns in net_namespace_list.to_list(net_type_symname, "list"):
            yield from net_ns.dev_base_head.to_list(net_device_symname, "dev_list")

    def _generator(self):
        for net_dev in self._enumerate_net_namespace_list():
            for (
                net_ns_id,
                iface_ifindex,
                iface_name,
                mac_addr,
                promisc,
                ip6_addr,
                prefix_len,
                scope_type,
                operational_state,
            ) in self._gather_net_dev_info(net_dev):
                yield (
                    0,
                    (
                        net_ns_id or renderers.NotAvailableValue(),
                        iface_ifindex,
                        iface_name,
                        mac_addr,
                        promisc,
                        ip6_addr,
                        prefix_len,
                        scope_type,
                        operational_state,
                    ),
                )

    def run(self):
        headers = [
            ("NetNS", int),
            ("Index", int),
            ("Interface", str),
            ("MAC", str),
            ("Promiscuous", bool),
            ("IP", str),
            ("Prefix", int),
            ("Scope Type", str),
            ("State", str),
        ]

        return renderers.TreeGrid(headers, self._generator())


class Link(plugins.PluginInterface):
    """Lists information about network interfaces similar to `ip link show`"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="Net", component=network.NetSymbols, version=(1, 0, 0)
            ),
        ]

    def _gather_net_dev_link_info(self, net_device):
        mac_addr = net_device.get_mac_address()
        operational_state = net_device.get_operational_state()
        iface_name = net_device.get_device_name()
        mtu = net_device.mtu
        qdisc_name = net_device.get_qdisc_name()
        qlen = net_device.get_queue_length()
        try:
            net_ns_id = net_device.get_net_namespace_id()
        except AttributeError:
            net_ns_id = renderers.NotAvailableValue()

        # Format flags to string. Drop IFF_ to match iproute2 'ip link' output.
        # Also, note that iproute2 removes IFF_RUNNING, see print_link_flags()
        flags_list = [
            flag.replace("IFF_", "")
            for flag in net_device.get_flag_names()
            if flag != "IFF_RUNNING"
        ]
        flags_str = ",".join(flags_list)

        yield (
            net_ns_id or renderers.NotAvailableValue(),
            iface_name,
            mac_addr,
            operational_state,
            mtu,
            qdisc_name or renderers.NotAvailableValue(),
            qlen,
            flags_str,
        )

    def _generator(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        network.NetSymbols.apply(self.context.symbol_space[vmlinux.symbol_table_name])

        net_type_symname = vmlinux.symbol_table_name + constants.BANG + "net"
        net_device_symname = vmlinux.symbol_table_name + constants.BANG + "net_device"

        # 'net_namespace_list' exists from kernels >= 2.6.24
        net_namespace_list = vmlinux.object_from_symbol("net_namespace_list")
        for net_ns in net_namespace_list.to_list(net_type_symname, "list"):
            for net_dev in net_ns.dev_base_head.to_list(net_device_symname, "dev_list"):
                for fields in self._gather_net_dev_link_info(net_dev):
                    yield 0, fields

    def run(self):
        headers = [
            ("NS", int),
            ("Interface", str),
            ("MAC", str),
            ("State", str),
            ("MTU", int),
            ("Qdisc", str),
            ("Qlen", int),
            ("Flags", str),
        ]

        return renderers.TreeGrid(headers, self._generator())



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""A `Context` maintains the accumulated state required for various plugins and
framework functions.

This has been made an object to allow quick swapping and changing of
contexts, to allow a plugin to act on multiple different contexts
without them interfering with each other.
"""

import functools
import hashlib
import logging
import re
from typing import Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

from volatility3.framework import constants, interfaces, symbols, exceptions
from volatility3.framework.objects import templates

vollog = logging.getLogger(__name__)


class Context(interfaces.context.ContextInterface):
    """Maintains the context within which to construct objects.

    The context object is the main method of carrying around state that's been constructed for the purposes of
    investigating memory.  It contains a symbol_space of all the symbols that can be accessed by plugins using the
    context.  It also contains the memory made up of data and translation layers, and it contains a factory method
    for creating new objects.

    Other context objects can be constructed as long as they support the
    :class:`~volatility3.framework.interfaces.context.ContextInterface`.  This is the primary context object to be used
    in the volatility framework.  It maintains the
    """

    def __init__(self) -> None:
        """Initializes the context."""
        super().__init__()
        self._symbol_space = symbols.SymbolSpace()
        self._module_space = ModuleCollection()
        self._memory = interfaces.layers.LayerContainer()
        self._config = interfaces.configuration.HierarchicalDict()

    # ## Symbol Space Functions

    @property
    def config(self) -> interfaces.configuration.HierarchicalDict:
        """Returns a mutable copy of the configuration, but does not allow the
        whole configuration to be altered."""
        return self._config

    @config.setter
    def config(self, value: interfaces.configuration.HierarchicalDict) -> None:
        if not isinstance(value, interfaces.configuration.HierarchicalDict):
            raise TypeError("Config must be of type HierarchicalDict")
        self._config = value

    @property
    def modules(self) -> interfaces.context.ModuleContainer:
        """A container for modules loaded in this context"""
        return self._module_space

    @property
    def symbol_space(self) -> interfaces.symbols.SymbolSpaceInterface:
        """The space of all symbols that can be accessed within this
        context."""
        return self._symbol_space

    @property
    def layers(self) -> interfaces.layers.LayerContainer:
        """A LayerContainer object, allowing access to all data and translation
        layers currently available within the context."""
        return self._memory

    # ## Translation Layer Functions

    def add_layer(self, layer: interfaces.layers.DataLayerInterface) -> None:
        """Adds a named translation layer to the context.

        Args:
            layer: The layer to be added to the memory

        Raises:
            volatility3.framework.exceptions.LayerException: if the layer is already present, or has
                unmet dependencies
        """
        self._memory.add_layer(layer)

    # ## Object Factory Functions

    def object(
        self,
        object_type: Union[str, interfaces.objects.Template],
        layer_name: str,
        offset: int,
        native_layer_name: Optional[str] = None,
        **arguments,
    ) -> interfaces.objects.ObjectInterface:
        """Object factory, takes a context, symbol, offset and optional
        layername.

        Looks up the layername in the context, finds the object template based on the symbol,
        and constructs an object using the object template on the layer at the offset.

        Args:
            object_type: The name (or template) of the symbol type on which to construct the object.  If this is a name, it should contain an explicit table name.
            layer_name: The name of the layer on which to construct the object
            offset: The offset within the layer at which the data used to create the object lives
            native_layer_name: The name of the layer the object references (for pointers) if different to layer_name

        Returns:
            A fully constructed object
        """
        if not isinstance(object_type, interfaces.objects.Template):
            try:
                object_template = self._symbol_space.get_type(object_type)
            except exceptions.SymbolError:
                object_template = self._symbol_space.get_enumeration(object_type)
        else:
            if isinstance(object_type, templates.ReferenceTemplate):
                object_type = self._symbol_space.get_type(object_type.vol.type_name)
            object_template = object_type
            # Ensure that if a pre-constructed type is provided we just instantiate it
            arguments.update(object_template.vol)

        object_template = object_template.clone()
        object_template.update_vol(**arguments)
        return object_template(
            context=self,
            object_info=interfaces.objects.ObjectInformation(
                layer_name=layer_name,
                offset=offset,
                native_layer_name=native_layer_name or layer_name,
                size=object_template.size,
            ),
        )

    def module(
        self,
        module_name: str,
        layer_name: str,
        offset: int,
        native_layer_name: Optional[str] = None,
        size: Optional[int] = None,
    ) -> interfaces.context.ModuleInterface:
        """Constructs a new os-independent module.

        Args:
            module_name: The name of the module
            layer_name: The layer within the context in which the module exists
            offset: The offset at which the module exists in the layer
            native_layer_name: The default native layer for objects constructed by the module
            size: The size, in bytes, that the module occupies from offset location within the layer named layer_name
        """
        if size:
            return SizedModule.create(
                self,
                module_name=module_name,
                layer_name=layer_name,
                offset=offset,
                size=size,
                native_layer_name=native_layer_name,
            )
        return Module.create(
            self,
            module_name=module_name,
            layer_name=layer_name,
            offset=offset,
            native_layer_name=native_layer_name,
        )


def get_module_wrapper(method: str) -> Callable:
    """Returns a symbol using the symbol_table_name of the Module."""

    def wrapper(self, name: str) -> Callable:
        if constants.BANG not in name:
            name = self.symbol_table_name + constants.BANG + name
        elif name.startswith(self.symbol_table_name + constants.BANG):
            pass
        else:
            raise ValueError(f"Cannot reference another module when calling {method}")
        return getattr(self._context.symbol_space, method)(name)

    for entry in [
        "__annotations__",
        "__doc__",
        "__module__",
        "__name__",
        "__qualname__",
    ]:
        proxy_interface = getattr(interfaces.context.ModuleInterface, method)
        if hasattr(proxy_interface, entry):
            setattr(wrapper, entry, getattr(proxy_interface, entry))

    return wrapper


class Module(interfaces.context.ModuleInterface):
    @classmethod
    def create(
        cls,
        context: interfaces.context.ContextInterface,
        module_name: str,
        layer_name: str,
        offset: int,
        **kwargs,
    ) -> "Module":
        pathjoin = interfaces.configuration.path_join
        # Check if config_path is None
        free_module_name = context.modules.free_module_name(module_name)
        config_path = kwargs.get("config_path", None)
        if config_path is None:
            config_path = pathjoin("temporary", "modules", free_module_name)
        # Populate the configuration
        context.config[pathjoin(config_path, "layer_name")] = layer_name
        context.config[pathjoin(config_path, "offset")] = offset
        # This is important, since the module_name may be changed in case it is already in use
        if "symbol_table_name" not in kwargs:
            kwargs["symbol_table_name"] = module_name
        for arg in kwargs:
            context.config[pathjoin(config_path, arg)] = kwargs.get(arg, None)
        # Construct the object
        return_val = cls(context, config_path, free_module_name)
        context.add_module(return_val)
        context.config[config_path] = return_val.name
        # Add the module to the context modules collection
        return return_val

    def object(
        self,
        object_type: str,
        offset: Optional[int] = None,
        native_layer_name: Optional[str] = None,
        absolute: bool = False,
        **kwargs,
    ) -> "interfaces.objects.ObjectInterface":
        """Returns an object created using the symbol_table_name and layer_name
        of the Module.

        Args:
            object_type: Name of the type/enumeration (within the module) to construct
            offset: The location of the object, ignored when symbol_type is SYMBOL
            native_layer_name: Name of the layer in which constructed objects are made (for pointers)
            absolute: whether the type's offset is absolute within memory or relative to the module
        """
        if constants.BANG not in object_type:
            object_type = self.symbol_table_name + constants.BANG + object_type
        elif not object_type.startswith(self.symbol_table_name + constants.BANG):
            raise ValueError(
                "Cannot reference another module when constructing an object"
            )

        if offset is None:
            raise TypeError("Offset must not be None for non-symbol objects")

        if not absolute:
            offset += self._offset

        # We have to allow using an alternative layer name due to pool scanners switching
        # to the memory layer for scanning samples prior to Windows 10.
        layer_name = kwargs.pop("layer_name", self._layer_name)

        return self._context.object(
            object_type=object_type,
            layer_name=layer_name,
            offset=offset,
            native_layer_name=native_layer_name or self._native_layer_name,
            **kwargs,
        )

    def object_from_symbol(
        self,
        symbol_name: str,
        native_layer_name: Optional[str] = None,
        absolute: bool = False,
        object_type: Optional[Union[str, "interfaces.objects.ObjectInterface"]] = None,
        **kwargs,
    ) -> "interfaces.objects.ObjectInterface":
        """Returns an object based on a specific symbol (containing type and
        offset information) and the layer_name of the Module.  This will throw
        a ValueError if the symbol does not contain an associated type, or if
        the symbol name is invalid.  It will throw a SymbolError if the symbol
        cannot be found.

        Args:
            symbol_name: Name of the symbol (within the module) to construct
            native_layer_name: Name of the layer in which constructed objects are made (for pointers)
            absolute: whether the symbol's address is absolute or relative to the module
            object_type: Override for the type from the symbol to use (or if the symbol type is missing)
        """
        if constants.BANG not in symbol_name:
            symbol_name = self.symbol_table_name + constants.BANG + symbol_name
        else:
            raise ValueError(
                "Cannot reference another module when constructing an object"
            )

        # Only set the offset if type is Symbol and we were given a name, not a template
        symbol_val = self._context.symbol_space.get_symbol(symbol_name)
        offset = symbol_val.address

        if not absolute:
            offset += self._offset

        if object_type is None:
            if symbol_val.type is None:
                raise TypeError(
                    f"Symbol {symbol_val.name} has no associated type and no object_type specified"
                )
            else:
                object_type = symbol_val.type

        # Ensure we don't use a layer_name other than the module's, why would anyone do that?
        if "layer_name" in kwargs:
            del kwargs["layer_name"]

        # Since type may be a template, we don't just call our own module method
        return self._context.object(
            object_type=object_type,
            layer_name=self._layer_name,
            offset=offset,
            native_layer_name=native_layer_name or self._native_layer_name,
            **kwargs,
        )

    def get_symbols_by_absolute_location(self, offset: int, size: int = 0) -> List[str]:
        """Returns the symbols within this module that live at the specified
        absolute offset provided."""
        if size < 0:
            raise ValueError("Size must be strictly non-negative")
        return list(
            self._context.symbol_space.get_symbols_by_location(
                offset=offset - self._offset,
                size=size,
                table_name=self.symbol_table_name,
            )
        )

    @property
    def symbols(self) -> Iterable[str]:
        return self.context.symbol_space[self.symbol_table_name].symbols

    get_symbol = get_module_wrapper("get_symbol")
    get_type = get_module_wrapper("get_type")
    get_enumeration = get_module_wrapper("get_enumeration")
    has_symbol = get_module_wrapper("has_symbol")
    has_type = get_module_wrapper("has_type")
    has_enumeration = get_module_wrapper("has_enumeration")


class SizedModule(Module):
    @property
    def size(self) -> int:
        """Returns the size of the module (0 for unknown size)"""
        size = self.config.get("size", 0)
        return size or 0

    @property  # type: ignore # FIXME: mypy #5107
    @functools.lru_cache
    def hash(self) -> str:
        """Hashes the module for equality checks.

        The mapping should be sorted and should be quicker than reading
        the data We turn it into JSON to make a common string and use a
        quick hash, because collisions are unlikely
        """
        layer = self._context.layers[self.layer_name]
        if not isinstance(layer, interfaces.layers.TranslationLayerInterface):
            raise TypeError("Hashing modules on non-TranslationLayers is not allowed")
        return hashlib.md5(
            bytes(
                str(list(layer.mapping(self.offset, self.size, ignore_errors=True))),
                "utf-8",
            )
        ).hexdigest()

    def get_symbols_by_absolute_location(self, offset: int, size: int = 0) -> List[str]:
        """Returns the symbols within this module that live at the specified
        absolute offset provided."""
        if offset > self._offset + self.size:
            return []
        return super().get_symbols_by_absolute_location(offset, size)


class ModuleCollection(interfaces.context.ModuleContainer):
    """Class to contain a collection of SizedModules and reason about their
    contents."""

    def __init__(self, modules: Optional[List[SizedModule]] = None) -> None:
        self._modules: Dict[str, SizedModule] = {}
        super().__init__(modules)

    def deduplicate(self) -> "ModuleCollection":
        """Returns a new deduplicated ModuleCollection featuring no repeated
        modules (based on data hash)

        All 0 sized modules will have identical hashes and are therefore
        included in the deduplicated version
        """
        new_modules = []
        seen: Set[str] = set()
        for mod in self._modules:
            if self._modules[mod].hash not in seen or self._modules[mod].size == 0:
                new_modules.append(mod)
                seen.add(self._modules[mod].hash)
        return ModuleCollection(new_modules)

    def free_module_name(self, prefix: str = "module") -> str:
        """Returns an unused module name"""
        existing_names = [name for name in self if re.match(rf"^{prefix}[0-9]*$", name)]
        if not existing_names:
            return prefix
        count = len(existing_names)
        while prefix + str(count) in self:
            count += 1
        return prefix + str(count)

    @property
    def modules(self) -> "ModuleCollection":
        """A name indexed dictionary of modules using that name in this
        collection."""
        vollog.warning(
            "This method has been deprecated in favour of the ModuleCollection acting as a dictionary itself"
        )
        return self

    def get_module_symbols_by_absolute_location(
        self, offset: int, size: int = 0
    ) -> Iterable[Tuple[str, List[str]]]:
        """Returns a tuple of (module_name, list_of_symbol_names) for each
        module, where symbols live at the absolute offset in memory
        provided."""
        if size < 0:
            raise ValueError("Size must be strictly non-negative")
        for module_name in self._modules:
            module = self._modules[module_name]
            if isinstance(module, SizedModule):
                if (offset <= module.offset + module.size) and (
                    offset + size >= module.offset
                ):
                    yield (
                        module.name,
                        module.get_symbols_by_absolute_location(offset, size),
                    )


class ConfigurableModule(Module, interfaces.configuration.ConfigurableInterface):
    def __init__(
        self, context: interfaces.context.ContextInterface, config_path: str, name: str
    ) -> None:
        interfaces.configuration.ConfigurableInterface.__init__(
            self, context, config_path
        )
        layer_name = self.config["layer_name"]
        offset = self.config["offset"]
        symbol_table_name = self.config["symbol_table_name"]
        interfaces.configuration.ConfigurableInterface.__init__(
            self, context, config_path
        )
        Module.__init__(
            self, context, name, layer_name, offset, symbol_table_name, layer_name
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from typing import List, Tuple, Optional
import logging
from volatility3.framework import interfaces
from volatility3.framework import renderers, symbols
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class Malfind(interfaces.plugins.PluginInterface):
    """Lists process memory ranges that potentially contain injected code."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 4)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.IntRequirement(
                name="dump-size",
                description="Amount of bytes to dump for each dirty region/page found - Default 64 bytes",
                optional=True,
                default=64,
            ),
            requirements.BooleanRequirement(
                name="dump-page",
                description="Dump each dirty page and content - Default off",
                optional=True,
                default=False,
            ),
        ]

    def _list_injections(
        self, task
    ) -> Tuple[interfaces.objects.ObjectInterface, Optional[str], bytes]:
        """Generate memory regions for a process that may contain injected
        code."""

        proc_layer_name = task.add_process_layer()
        if not proc_layer_name:
            return None

        proc_layer = self.context.layers[proc_layer_name]

        dump_size = self.config["dump-size"]

        # Dumping page defaults to off, as in case a whole r-xp region is dirty
        # this would likely dump 1000's of pages which might not always be wise nor necessary

        dump_page = self.config["dump-page"]

        for vma in task.mm.get_vma_iter():
            vma_name = vma.get_name(self.context, task)
            vollog.debug(
                f"Injections : processing PID {task.pid} : VMA {vma_name} : {hex(vma.vm_start)}-{hex(vma.vm_end)}"
            )

            # If is_suspicious returns true, this means at least one page
            # in the region is dirty. If dump_page is true, then we dump
            # all dirty pages

            if vma.is_suspicious(proc_layer) and vma_name != "[vdso]":
                malicious_pages = vma.get_malicious_pages(proc_layer)
                offset = 0
                if dump_page:
                    # Dumping each dirty page
                    for page_addr in malicious_pages:
                        offset = page_addr - vma.vm_start
                        data = proc_layer.read(page_addr, dump_size, pad=True)
                        yield (
                            vma,
                            f"{vma_name}, page address: {page_addr:#x}, offset: {offset:#x}",
                            data,
                            offset,
                        )
                else:
                    # Original behaviour - Dump the start of the region (not necessarily matching the dirty page)
                    data = proc_layer.read(vma.vm_start, dump_size, pad=True)
                    yield vma, vma_name, data, offset

    def _generator(self, tasks):
        # determine if we're on a 32 or 64 bit kernel
        vmlinux = self.context.modules[self.config["kernel"]]
        is_32bit_arch = not symbols.symbol_table_is_64bit(
            context=self.context, symbol_table_name=vmlinux.symbol_table_name
        )

        for task in tasks:
            process_name = utility.array_to_string(task.comm)

            for vma, vma_name, data, offset in self._list_injections(task):
                if is_32bit_arch:
                    architecture = "intel"
                else:
                    architecture = "intel64"

                disasm = renderers.Disassembly(
                    data, vma.vm_start + offset, architecture
                )

                yield (
                    0,
                    (
                        task.pid,
                        process_name,
                        format_hints.Hex(vma.vm_start),
                        format_hints.Hex(vma.vm_end),
                        vma_name or renderers.NotAvailableValue(),
                        vma.get_protection(),
                        format_hints.HexBytes(data),
                        disasm,
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Start", format_hints.Hex),
                ("End", format_hints.Hex),
                ("Path", str),
                ("Protection", str),
                ("Hexdump", format_hints.HexBytes),
                ("Disasm", renderers.Disassembly),
            ],
            self._generator(
                pslist.PsList.list_tasks(
                    self.context, self.config["kernel"], filter_func=filter_func
                )
            ),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from typing import List

from volatility3.framework import renderers, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.symbols import linux
from volatility3.framework.renderers import format_hints


class VMCoreInfo(plugins.PluginInterface):
    """Enumerate VMCoreInfo tables"""

    _required_framework_version = (2, 11, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.TranslationLayerRequirement(
                name="primary", description="Memory layer to scan"
            ),
            requirements.VersionRequirement(
                name="VMCoreInfo", component=linux.VMCoreInfo, version=(1, 0, 0)
            ),
        ]

    def _generator(self):
        layer_name = self.config["primary"]
        for (
            vmcoreinfo_offset,
            vmcoreinfo,
        ) in linux.VMCoreInfo.search_vmcoreinfo_elf_note(
            context=self.context,
            layer_name=layer_name,
        ):
            for key, value in vmcoreinfo.items():
                if key.startswith("SYMBOL(") or key == "KERNELOFFSET":
                    value = hex(value)
                else:
                    value = str(value)

                yield 0, (format_hints.Hex(vmcoreinfo_offset), key, value)

    def run(self):
        headers = [
            ("Offset", format_hints.Hex),
            ("Key", str),
            ("Value", str),
        ]
        return renderers.TreeGrid(headers, self._generator())



# This file is Copyright 2021 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from collections import namedtuple
from typing import Tuple, List, Iterable, Union

from volatility3.framework import renderers, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.symbols import linux
from volatility3.plugins.linux import pslist


vollog = logging.getLogger(__name__)

MountInfoData = namedtuple(
    "MountInfoData",
    (
        "mnt_id",
        "parent_id",
        "st_dev",
        "mnt_root_path",
        "path_root",
        "mnt_opts",
        "fields",
        "mnt_type",
        "devname",
        "sb_opts",
    ),
)


class MountInfo(plugins.PluginInterface):
    """Lists mount points on processes mount namespaces"""

    _required_framework_version = (2, 2, 0)
    _version = (1, 2, 4)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.VersionRequirement(
                name="linuxutils", component=linux.LinuxUtilities, version=(2, 1, 0)
            ),
            requirements.ListRequirement(
                name="pids",
                description="Filter on specific process IDs.",
                element_type=int,
                optional=True,
            ),
            requirements.ListRequirement(
                name="mntns",
                description="Filter results by mount namespace. "
                "Otherwise, all of them are shown.",
                element_type=int,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="mount-format",
                description="Shows a brief summary of the mount points information "
                "with similar output format to the older /proc/[pid]/mounts or the "
                "user-land command 'mount -l'.",
                optional=True,
                default=False,
            ),
        ]

    @classmethod
    def get_mountinfo(
        cls, mnt, task
    ) -> Union[
        None, Tuple[int, int, str, str, str, List[str], List[str], str, str, List[str]]
    ]:
        """Extract various information about a mount point.
        It mimics the Linux kernel show_mountinfo function.
        """
        mnt_root = mnt.get_mnt_root()
        if not mnt_root:
            return None

        path_root = linux.LinuxUtilities.get_path_mnt(task, mnt)
        if not path_root:
            return None

        mnt_root_path = mnt_root.path()

        mnt_id: int = mnt.mnt_id
        parent_id: int = mnt.mnt_parent.mnt_id

        superblock = mnt.get_mnt_sb()
        if not (superblock and superblock.is_readable()):
            return None

        st_dev = f"{superblock.major}:{superblock.minor}"

        mnt_opts: List[str] = []
        mnt_opts.append(mnt.get_flags_access())
        mnt_opts.extend(mnt.get_flags_opts())

        # Tagged fields
        fields: List[str] = []
        if mnt.is_shared():
            fields.append(f"shared:{mnt.mnt_group_id}")

        if mnt.is_slave():
            master = mnt.mnt_master.mnt_group_id
            fields.append(f"master:{master}")
            dominating_id = mnt.get_dominating_id(task.fs.root)
            if dominating_id and dominating_id != master:
                fields.append(f"propagate_from:{dominating_id}")

        if mnt.is_unbindable():
            fields.append("unbindable")

        mnt_type = superblock.get_type()

        devname = mnt.get_devname()
        if not devname:
            devname = "none"

        sb_opts: List[str] = []
        sb_opts.append(superblock.get_flags_access())
        sb_opts.extend(superblock.get_flags_opts())

        return MountInfoData(
            mnt_id,
            parent_id,
            st_dev,
            mnt_root_path,
            path_root,
            mnt_opts,
            fields,
            mnt_type,
            devname,
            sb_opts,
        )

    @staticmethod
    def _get_tasks_mountpoints(
        tasks: Iterable[interfaces.objects.ObjectInterface],
        filtered_by_pids: bool = False,
    ):
        seen_mountpoints = set()
        for task in tasks:
            if not (
                task
                and task.fs
                and task.fs.is_readable()
                and task.nsproxy
                and task.nsproxy.is_readable()
                and task.nsproxy.mnt_ns
                and task.nsproxy.mnt_ns.is_readable()
            ):
                # This task doesn't have all the information required.
                # It should be a kernel < 2.6.30
                continue

            mnt_namespace = task.nsproxy.mnt_ns
            try:
                mnt_ns_id = mnt_namespace.get_inode()
            except AttributeError:
                mnt_ns_id = renderers.NotAvailableValue()

            for mount in mnt_namespace.get_mount_points():
                # When PIDs are filtered, it makes sense that the user want to
                # see each of those processes mount points. So we don't filter
                # by mount id in this case.
                if not filtered_by_pids:
                    mnt_id = int(mount.mnt_id)
                    if mnt_id in seen_mountpoints:
                        continue
                    else:
                        seen_mountpoints.add(mnt_id)

                yield task, mount, mnt_ns_id

    def _generator(
        self,
        tasks: Iterable[interfaces.objects.ObjectInterface],
        mnt_ns_ids: List[int],
        mount_format: bool = False,
        filtered_by_pids: bool = False,
    ) -> Iterable[Tuple[int, Tuple]]:
        show_filter_warning = False
        for task, mnt, mnt_ns_id in self._get_tasks_mountpoints(
            tasks, filtered_by_pids
        ):
            if mnt_ns_ids and isinstance(mnt_ns_id, renderers.NotAvailableValue):
                show_filter_warning = True

            if (
                not isinstance(mnt_ns_id, renderers.NotAvailableValue)
                and mnt_ns_ids
                and mnt_ns_id not in mnt_ns_ids
            ):
                continue

            mnt_info = self.get_mountinfo(mnt, task)
            if mnt_info is None:
                continue

            if mount_format:
                all_opts = set()
                all_opts.update(mnt_info.mnt_opts)
                all_opts.update(mnt_info.sb_opts)
                all_opts_str = ",".join(all_opts)

                extra_fields_values = [
                    mnt_info.devname,
                    mnt_info.path_root,
                    mnt_info.mnt_type,
                    all_opts_str,
                ]
            else:
                mnt_opts_str = ",".join(mnt_info.mnt_opts)
                fields_str = " ".join(mnt_info.fields)
                sb_opts_str = ",".join(mnt_info.sb_opts)

                extra_fields_values = [
                    mnt_info.mnt_id,
                    mnt_info.parent_id,
                    mnt_info.st_dev,
                    mnt_info.mnt_root_path,
                    mnt_info.path_root,
                    mnt_opts_str,
                    fields_str,
                    mnt_info.mnt_type,
                    mnt_info.devname,
                    sb_opts_str,
                ]

            fields_values = [mnt_ns_id]
            if filtered_by_pids:
                fields_values.append(task.pid)
            fields_values.extend(extra_fields_values)

            yield (0, fields_values)

        if show_filter_warning:
            vollog.warning(
                "Could not filter by mount namespace id. This field is not available in this kernel."
            )

    @classmethod
    def get_superblocks(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Yield file system superblocks based on the task's mounted filesystems.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            vmlinux_module_name: The name of the kernel module on which to operate

        Yields:
            super_block: Kernel's struct super_block object
        """
        # No filter so that we get all the mount namespaces from all tasks
        tasks = pslist.PsList.list_tasks(context, vmlinux_module_name)

        seen_sb_ptr = set()
        for task, mnt, _mnt_ns_id in cls._get_tasks_mountpoints(tasks):
            path_root = linux.LinuxUtilities.get_path_mnt(task, mnt)
            if not path_root:
                continue

            sb_ptr = mnt.get_mnt_sb()
            if not (sb_ptr and sb_ptr.is_readable()):
                continue

            if sb_ptr in seen_sb_ptr:
                continue
            seen_sb_ptr.add(int(sb_ptr))

            superblock = sb_ptr.dereference()

            yield superblock, path_root

    def run(self):
        pids = self.config.get("pids")
        mount_ns_ids = self.config.get("mntns")
        mount_format = self.config.get("mount-format")

        pid_filter = pslist.PsList.create_pid_filter(pids)
        tasks = pslist.PsList.list_tasks(
            self.context, self.config["kernel"], filter_func=pid_filter
        )

        columns = [("MNT_NS_ID", int)]
        # The PID column does not make sense when a PID filter is not specified. In that case, the default behavior is
        # to displays the mountpoints per namespace.
        if pids:
            columns.append(("PID", int))
            filtered_by_pids = True
        else:
            filtered_by_pids = False

        if self.config.get("mount-format"):
            extra_columns = [
                ("DEVNAME", str),
                ("PATH", str),
                ("FSTYPE", str),
                ("MNT_OPTS", str),
            ]
        else:
            # /proc/[pid]/mountinfo output format
            extra_columns = [
                ("MOUNT ID", int),
                ("PARENT_ID", int),
                ("MAJOR:MINOR", str),
                ("ROOT", str),
                ("MOUNT_POINT", str),
                ("MOUNT_OPTIONS", str),
                ("FIELDS", str),
                ("FSTYPE", str),
                ("MOUNT_SRC", str),
                ("SB_OPTIONS", str),
            ]

        columns.extend(extra_columns)

        return renderers.TreeGrid(
            columns,
            self._generator(tasks, mount_ns_ids, mount_format, filtered_by_pids),
        )



# This file is Copyright 2022 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import datetime
import logging
from typing import Iterator, NamedTuple, Optional, Tuple, Union

from volatility3.framework import constants, exceptions, interfaces, objects, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import conversion, format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows.extensions import mft
from volatility3.plugins import timeliner, yarascan

vollog = logging.getLogger(__name__)


class MFTScan(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Scans for MFT FILE objects present in a particular windows memory image."""

    _required_framework_version = (2, 26, 0)

    _version = (3, 0, 0)

    class MFTScanResult(NamedTuple):
        offset: format_hints.Hex
        record_type: str
        record_number: objects.Integer
        link_count: objects.Integer
        mft_type: str
        permissions: Union[str, interfaces.renderers.BaseAbsentValue]
        attribute_type: str
        created: Union[interfaces.renderers.BaseAbsentValue, datetime.datetime]
        modified: Union[interfaces.renderers.BaseAbsentValue, datetime.datetime]
        updated: Union[interfaces.renderers.BaseAbsentValue, datetime.datetime]
        accessed: Union[interfaces.renderers.BaseAbsentValue, datetime.datetime]
        filename: Union[interfaces.renderers.BaseAbsentValue, objects.String]

    @classmethod
    def get_requirements(cls):
        return [
            requirements.TranslationLayerRequirement(
                name="primary",
                description="Memory layer for the kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="yarascanner", component=yarascan.YaraScanner, version=(2, 1, 0)
            ),
            requirements.VersionRequirement(
                name="yarascan", component=yarascan.YaraScan, version=(2, 0, 0)
            ),
        ]

    @classmethod
    def enumerate_mft_records(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        primary_layer_name: str,
    ) -> Iterator[mft.MFTEntry]:
        try:
            primary = context.layers[primary_layer_name]
        except KeyError:
            vollog.error(
                "Unable to obtain primary layer for scanning. Please file a bug on GitHub about this issue."
            )
            return

        try:
            memory_layer_name = primary.config["memory_layer"]
        except KeyError:
            vollog.error(
                "Unable to obtain memory layer from primary layer. Please file a bug on GitHub about this issue."
            )
            return

        layer = context.layers[memory_layer_name]

        # Yara Rule to scan for MFT Header Signatures
        rules = yarascan.YaraScan.process_yara_options(
            {"yara_string": "/FILE0|FILE\\*|BAAD/"}
        )

        # Read in the Symbol File
        symbol_table_name = intermed.IntermediateSymbolTable.create(
            context=context,
            config_path=config_path,
            sub_path="windows",
            filename="mft",
            class_types={
                "FILE_NAME_ENTRY": mft.MFTFileName,
                "MFT_ENTRY": mft.MFTEntry,
                "ATTRIBUTE": mft.MFTAttribute,
            },
        )

        # get each of the individual Field Sets

        mft_object_type_name = symbol_table_name + constants.BANG + "MFT_ENTRY"

        # Scan the layer for Raw MFT records and parse the fields
        for offset, _rule_name, _name, _value in layer.scan(
            context=context, scanner=yarascan.YaraScanner(rules=rules)
        ):
            mft_record: mft.MFTEntry = context.object(
                mft_object_type_name,
                offset=offset,
                layer_name=layer.name,
            )

            yield mft_record

    @classmethod
    def parse_standard_information_records(
        cls, mft_record: mft.MFTEntry
    ) -> Iterator[Tuple[int, MFTScanResult]]:
        # MFT Flags determine the file type or dir
        # If we don't have a valid enum, coerce to hex so we can keep the record
        try:
            mft_flag = mft_record.Flags.lookup()
        except ValueError:
            mft_flag = hex(mft_record.Flags)

        # Standard Information Attribute
        try:
            # There should only be one STANDARD_INFORMATION attribute, but we
            # do this just in case.
            for std_information in mft_record.standard_information_entries():
                yield (
                    0,
                    cls.MFTScanResult(
                        format_hints.Hex(std_information.vol.offset),
                        str(mft_record.get_signature()),
                        mft_record.RecordNumber,
                        mft_record.LinkCount,
                        mft_flag,
                        renderers.NotApplicableValue(),
                        "STANDARD_INFORMATION",
                        conversion.wintime_to_datetime(std_information.CreationTime),
                        conversion.wintime_to_datetime(std_information.ModifiedTime),
                        conversion.wintime_to_datetime(std_information.UpdatedTime),
                        conversion.wintime_to_datetime(std_information.AccessedTime),
                        renderers.NotApplicableValue(),
                    ),
                )
        except exceptions.InvalidAddressException:
            pass

    @classmethod
    def parse_filename_records(
        cls, mft_record: mft.MFTEntry
    ) -> Iterator[Tuple[int, MFTScanResult]]:
        # MFT Flags determine the file type or dir
        # If we don't have a valid enum, coerce to hex so we can keep the record
        try:
            mft_flag = mft_record.Flags.lookup()
        except ValueError:
            mft_flag = hex(mft_record.Flags)

        # File Name Attribute
        try:
            for filename_info in mft_record.filename_entries():
                # If we don't have a valid enum, coerce to hex so we can keep the record
                try:
                    permissions = filename_info.Flags.lookup()
                except ValueError:
                    permissions = hex(filename_info.Flags)

                yield (
                    1,
                    cls.MFTScanResult(
                        format_hints.Hex(filename_info.vol.offset),
                        str(mft_record.get_signature()),
                        mft_record.RecordNumber,
                        mft_record.LinkCount,
                        mft_flag,
                        permissions,
                        "FILE_NAME",
                        conversion.wintime_to_datetime(filename_info.CreationTime),
                        conversion.wintime_to_datetime(filename_info.ModifiedTime),
                        conversion.wintime_to_datetime(filename_info.UpdatedTime),
                        conversion.wintime_to_datetime(filename_info.AccessedTime),
                        filename_info.get_full_name(),
                    ),
                )
        except exceptions.InvalidAddressException:
            return

    @classmethod
    def parse_mft_records(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        primary_layer_name: str,
    ) -> Iterator[Tuple[int, MFTScanResult]]:
        for mft_record in cls.enumerate_mft_records(
            context=context,
            config_path=config_path,
            primary_layer_name=primary_layer_name,
        ):
            yield from cls.parse_standard_information_records(mft_record)
            yield from cls.parse_filename_records(mft_record)

    def _generator(self):
        for level, record in self.parse_mft_records(
            self.context,
            self.config_path,
            self.config["primary"],
        ):
            # Convert all `objects.PrimitiveObject` to their simpler Python
            # types. This is normally not something we would do, since it's
            # lossy and prevents users from getting back to the data source,
            # but in this case memory usage is so extreme due to the number of
            # records that it becomes necessary. The rich types are still
            # exposed through classmethods.
            yield (
                level,
                (
                    record.offset,
                    record.record_type,
                    int(record.record_number),
                    int(record.link_count),
                    record.mft_type,
                    record.permissions,
                    record.attribute_type,
                    record.created,
                    record.modified,
                    record.updated,
                    record.accessed,
                    (
                        str(record.filename)
                        if isinstance(record.filename, objects.String)
                        else record.filename
                    ),
                ),
            )

    def generate_timeline(self):
        for record in self.enumerate_mft_records(
            self.context, self.config_path, self.config["primary"]
        ):
            fname = record.longest_filename()

            for _, item in self.parse_standard_information_records(record):
                description = f"MFT {item.attribute_type} entry for {fname}"
                yield (description, timeliner.TimeLinerType.CREATED, item.created)
                yield (description, timeliner.TimeLinerType.MODIFIED, item.modified)
                yield (description, timeliner.TimeLinerType.CHANGED, item.updated)
                yield (description, timeliner.TimeLinerType.ACCESSED, item.accessed)

            for _, item in self.parse_filename_records(record):
                description = f"MFT {item.attribute_type} entry for {item.filename}"
                yield (description, timeliner.TimeLinerType.CREATED, item.created)
                yield (description, timeliner.TimeLinerType.MODIFIED, item.modified)
                yield (description, timeliner.TimeLinerType.CHANGED, item.updated)
                yield (description, timeliner.TimeLinerType.ACCESSED, item.accessed)

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Record Type", str),
                ("Record Number", int),
                ("Link Count", int),
                ("MFT Type", str),
                ("Permissions", str),
                ("Attribute Type", str),
                ("Created", datetime.datetime),
                ("Modified", datetime.datetime),
                ("Updated", datetime.datetime),
                ("Accessed", datetime.datetime),
                ("Filename", str),
            ],
            self._generator(),
        )


class ADS(interfaces.plugins.PluginInterface):
    """Scans for Alternate Data Stream"""

    _required_framework_version = (2, 26, 0)

    _version = (2, 0, 0)

    class ADSResult(NamedTuple):
        offset: format_hints.Hex
        signature: objects.String
        record_number: objects.Integer
        attribute_type: str
        filename: Union[objects.String, interfaces.renderers.BaseAbsentValue]
        stream_name: Union[objects.String, interfaces.renderers.BaseAbsentValue]
        content: Union[renderers.LayerData, interfaces.renderers.BaseAbsentValue]

    @classmethod
    def get_requirements(cls):
        return [
            requirements.VersionRequirement(
                name="MFTScan", component=MFTScan, version=(3, 0, 0)
            ),
            requirements.TranslationLayerRequirement(
                name="primary",
                description="Memory layer for the kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="yarascanner", component=yarascan.YaraScanner, version=(2, 0, 0)
            ),
        ]

    @classmethod
    def parse_ads_data_records(cls, mft_record: mft.MFTEntry) -> Iterator[ADSResult]:
        for data_attr in mft_record.alternate_data_streams():
            record_filename = (
                mft_record.longest_filename() or renderers.NotAvailableValue()
            )
            content_obj = data_attr.get_resident_filecontent()
            content = (
                renderers.LayerData.from_object(content_obj)
                if content_obj
                else renderers.NotAvailableValue()
            )
            ads_filename = (
                data_attr.get_resident_filename() or renderers.NotAvailableValue()
            )

            yield cls.ADSResult(
                format_hints.Hex(data_attr.Attr_Data.vol.offset),
                mft_record.get_signature(),
                mft_record.RecordNumber,
                data_attr.Attr_Header.AttrType.lookup(),
                record_filename,
                ads_filename,
                content,
            )

    def _generator(self):
        for mft_entry in MFTScan.enumerate_mft_records(
            self.context,
            self.config_path,
            self.config["primary"],
        ):
            for record in self.parse_ads_data_records(mft_entry):
                # Convert all `objects.PrimitiveObject` to their simpler Python
                # types. This is normally not something we would do, since it's
                # lossy and prevents users from getting back to the data source,
                # but in this case memory usage is so extreme due to the number of
                # records that it becomes necessary. The rich types are still
                # exposed through classmethods.
                yield (
                    0,
                    (
                        record.offset,
                        str(record.signature),
                        int(record.record_number),
                        record.attribute_type,
                        (
                            str(record.filename)
                            if isinstance(record.filename, objects.String)
                            else record.filename
                        ),
                        (
                            str(record.stream_name)
                            if isinstance(record.stream_name, objects.String)
                            else record.stream_name
                        ),
                        record.content,
                    ),
                )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Record Type", str),
                ("Record Number", int),
                ("MFT Type", str),
                ("Filename", str),
                ("ADS Filename", str),
                ("Hexdump", renderers.LayerData),
            ],
            self._generator(),
        )


class ResidentData(interfaces.plugins.PluginInterface):
    """Scans for MFT Records with Resident Data"""

    _required_framework_version = (2, 26, 0)

    _version = (2, 0, 0)

    class ResidentDataResult(NamedTuple):
        offset: format_hints.Hex
        signature: objects.String
        record_number: int
        attribute_type: str
        filename: Union[objects.String, interfaces.renderers.BaseAbsentValue]
        content: Union[renderers.LayerData, interfaces.renderers.BaseAbsentValue]

    @classmethod
    def get_requirements(cls):
        return [
            requirements.VersionRequirement(
                name="MFTScan", component=MFTScan, version=(3, 0, 0)
            ),
            requirements.TranslationLayerRequirement(
                name="primary",
                description="Memory layer for the kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="yarascanner", component=yarascan.YaraScanner, version=(2, 0, 0)
            ),
        ]

    @classmethod
    def parse_resident_data(
        cls,
        mft_record: mft.MFTEntry,
    ) -> Optional[ResidentDataResult]:
        """
        Returns the parsed data from a MFT record
        """

        try:
            attr = next(mft_record.resident_data_attributes())
        except StopIteration:
            return None

        content = attr.get_resident_filecontent()
        if content:
            content = renderers.LayerData.from_object(content)
        else:
            content = renderers.NotAvailableValue()

        # Choose the longest of the two, since it often includes a DOS 8.3 name
        filename = mft_record.longest_filename() or renderers.NotAvailableValue()

        return cls.ResidentDataResult(
            format_hints.Hex(attr.Attr_Data.vol.offset),
            mft_record.get_signature(),
            mft_record.RecordNumber,
            attr.Attr_Header.AttrType.lookup(),
            filename,
            content,
        )

    def _generator(self):
        for mft_record in MFTScan.enumerate_mft_records(
            self.context,
            self.config_path,
            self.config["primary"],
        ):
            resident_data_entry = self.parse_resident_data(mft_record)
            if resident_data_entry:
                # Convert all `objects.PrimitiveObject` to their simpler Python
                # types. This is normally not something we would do, since it's
                # lossy and prevents users from getting back to the data source,
                # but in this case memory usage is so extreme due to the number of
                # records that it becomes necessary. The rich types are still
                # exposed through classmethods.
                yield (
                    0,
                    (
                        resident_data_entry.offset,
                        str(resident_data_entry.signature),
                        int(resident_data_entry.record_number),
                        resident_data_entry.attribute_type,
                        str(resident_data_entry.filename),
                        resident_data_entry.content,
                    ),
                )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Record Type", str),
                ("Record Number", int),
                ("MFT Type", str),
                ("Filename", str),
                ("Hexdump", renderers.LayerData),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from volatility3.framework import exceptions
from volatility3.framework import objects
from volatility3.framework.objects import utility
from volatility3.framework.renderers import conversion


class hist_entry(objects.StructType):
    def is_valid(self):
        try:
            cmd = self.get_command()
            ts = utility.array_to_string(self.timestamp.dereference())
        except exceptions.InvalidAddressException:
            return False

        if not cmd or len(cmd) == 0:
            return False

        if not ts or len(ts) == 0:
            return False

        # At this point in time, the epoc integer size will
        # never be less than 10 characters, and the stamp is
        # always preceded by a pound/hash character.
        if len(ts) < 10 or str(ts)[0] != "#":
            return False

        # The final check is to make sure the entire string
        # is composed of numbers. Try to convert to an int.
        try:
            int(str(ts)[1:])
        except ValueError:
            return False

        return True

    def get_time_as_integer(self):
        # Get the string and remove the leading "#" from the timestamp
        time_string = utility.array_to_string(self.timestamp.dereference())[1:]
        # Convert the string into an integer (number of seconds)
        return int(time_string)

    def get_time_object(self):
        nsecs = self.get_time_as_integer()
        # Build a timestamp object from the integer
        return conversion.unixtime_to_datetime(nsecs)

    def get_command(self):
        return utility.array_to_string(self.line.dereference())



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import math
from typing import Optional, Dict, Any, List, Iterable, Tuple

from volatility3.framework import interfaces, constants, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import linear
from volatility3.framework.objects import utility
from volatility3.framework.symbols import intermed


class PDBFormatException(exceptions.LayerException):
    """Thrown when an error occurs with the underlying MSF file format."""


class PdbMultiStreamFormat(linear.LinearlyMappedLayer):
    _headers = {
        "MSF_HDR": "Microsoft C/C++ program database 2.00\r\n\x1a\x4a\x47",
        "BIG_MSF_HDR": "Microsoft C/C++ MSF 7.00\r\n\x1a\x44\x53",
    }

    def __init__(
        self,
        context: "interfaces.context.ContextInterface",
        config_path: str,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(context, config_path, name, metadata)
        self._base_layer = self.config["base_layer"]

        self._pdb_symbol_table = intermed.IntermediateSymbolTable.create(
            context, self._config_path, "windows", "pdb"
        )
        response = self._check_header()
        if response is None:
            raise PDBFormatException(name, "Could not find a suitable header")
        self._version, self._header = response
        self._streams: Dict[int, str] = {}

    @property
    def pdb_symbol_table(self) -> str:
        return self._pdb_symbol_table

    def read_streams(self):
        # Shortcut in case they've already been read
        if self._streams:
            return None

        # Recover the root table, by recovering the root table index table...
        module = self.context.module(self.pdb_symbol_table, self._base_layer, offset=0)
        entry_size = module.get_type("unsigned long").size

        root_table_num_pages = math.ceil(
            self._header.StreamInfo.StreamInfoSize / self._header.PageSize
        )
        root_index_size = math.ceil(
            (root_table_num_pages * entry_size) / self._header.PageSize
        )
        root_index = module.object(
            object_type="array",
            offset=self._header.vol.size,
            count=root_index_size,
            subtype=module.get_type("unsigned long"),
        )
        root_index_layer_name = self.create_stream_from_pages(
            "root_index",
            self._header.StreamInfo.StreamInfoSize,
            [x for x in root_index],
        )

        module = self.context.module(
            self.pdb_symbol_table, root_index_layer_name, offset=0
        )
        root_pages = module.object(
            object_type="array",
            offset=0,
            count=root_table_num_pages,
            subtype=module.get_type("unsigned long"),
        )
        root_layer_name = self.create_stream_from_pages(
            "root", self._header.StreamInfo.StreamInfoSize, [x for x in root_pages]
        )

        module = self.context.module(self.pdb_symbol_table, root_layer_name, offset=0)
        num_streams = module.object(object_type="unsigned long", offset=0)
        stream_sizes = module.object(
            object_type="array",
            offset=entry_size,
            count=num_streams,
            subtype=module.get_type("unsigned long"),
        )

        current_offset = (num_streams + 1) * entry_size

        for stream in range(num_streams):
            list_size = math.ceil(stream_sizes[stream] / self.page_size)
            if list_size == 0 or stream_sizes[stream] == 0xFFFFFFFF:
                self._streams[stream] = None
            else:
                stream_page_list = module.object(
                    object_type="array",
                    offset=current_offset,
                    count=list_size,
                    subtype=module.get_type("unsigned long"),
                )
                current_offset += list_size * entry_size
                self._streams[stream] = self.create_stream_from_pages(
                    "stream" + str(stream),
                    stream_sizes[stream],
                    [x for x in stream_page_list],
                )

    def create_stream_from_pages(
        self, stream_name: str, maximum_size: int, pages: List[int]
    ) -> str:
        # Construct a root layer based on a number of pages
        layer_name = self.name + "_" + stream_name
        path_join = interfaces.configuration.path_join
        config_path = path_join(self.config_path, stream_name)
        self.context.config[path_join(config_path, "base_layer")] = self.name
        self.context.config[path_join(config_path, "pages")] = pages
        self.context.config[path_join(config_path, "maximum_size")] = maximum_size
        layer = PdbMSFStream(self.context, config_path, layer_name)
        self.context.layers.add_layer(layer)
        return layer_name

    def _check_header(self) -> Optional[Tuple[str, interfaces.objects.ObjectInterface]]:
        """Verifies the header of the PDB file and returns the version of the
        file."""
        for header in self._headers:
            header_type = self.pdb_symbol_table + constants.BANG + header
            current_header = self.context.object(header_type, self._base_layer, 0)
            if utility.array_to_string(current_header.Magic) == self._headers[header]:
                if not (
                    current_header.PageSize < 0x100
                    or current_header.PageSize > (128 * 0x10000)
                ):
                    return header, current_header
        return None

    @property
    def page_size(self):
        return self._header.PageSize

    @property
    def dependencies(self) -> List[str]:
        """Returns a list of the lower layers that this layer is dependent
        upon."""
        return [self._base_layer]

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.TranslationLayerRequirement(name="base_layer", optional=False)
        ]

    @property
    def maximum_address(self) -> int:
        return self.context.layers[self._base_layer].maximum_address

    @property
    def minimum_address(self) -> int:
        return self.context.layers[self._base_layer].minimum_address

    def is_valid(self, offset: int, length: int = 1) -> bool:
        return self.context.layers[self._base_layer].is_valid(offset, length)

    def mapping(
        self, offset: int, length: int, ignore_errors: bool = False
    ) -> Iterable[Tuple[int, int, int, int, str]]:
        yield offset, length, offset, length, self._base_layer

    def get_stream(self, index) -> Optional["PdbMSFStream"]:
        self.read_streams()
        if index not in self._streams:
            raise PDBFormatException(self.name, "Stream not present")
        if self._streams[index]:
            layer = self.context.layers[self._streams[index]]
            if isinstance(layer, PdbMSFStream):
                return layer
        return None


class PdbMSFStream(linear.LinearlyMappedLayer):
    def __init__(
        self,
        context: "interfaces.context.ContextInterface",
        config_path: str,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(context, config_path, name, metadata)
        self._base_layer = self.config["base_layer"]
        self._pages = self.config.get("pages", [])
        self._pages_len = len(self._pages)
        if not self._pages:
            raise PDBFormatException(name, "Invalid/no pages specified")
        if not isinstance(self._pdb_layer, PdbMultiStreamFormat):
            raise TypeError("Base Layer must be a PdbMultiStreamFormat layer")

    @property
    def pdb_symbol_table(self) -> Optional[str]:
        layer = self._context.layers[self._base_layer]
        if isinstance(layer, PdbMultiStreamFormat):
            return layer.pdb_symbol_table
        else:
            return None

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ListRequirement(
                name="pages", element_type=int, min_elements=1
            ),
            requirements.TranslationLayerRequirement(name="base_layer"),
            requirements.IntRequirement(name="maximum_size"),
        ]

    def mapping(
        self, offset: int, length: int, ignore_errors: bool = False
    ) -> Iterable[Tuple[int, int, int, int, str]]:
        returned = 0
        page_size = self._pdb_layer.page_size
        while length > 0:
            page = (offset + returned) // page_size
            page_position = (offset + returned) % page_size
            chunk_size = min(page_size - page_position, length)
            if page >= self._pages_len:
                if not ignore_errors:
                    raise exceptions.InvalidAddressException(
                        layer_name=self.name, invalid_address=offset + returned
                    )
            else:
                yield (
                    offset + returned,
                    chunk_size,
                    (self._pages[page] * page_size) + page_position,
                    chunk_size,
                    self._base_layer,
                )
            returned += chunk_size
            length -= chunk_size

    @property
    def dependencies(self) -> List[str]:
        return [self._base_layer]

    def is_valid(self, offset: int, length: int = 1) -> bool:
        return self.context.layers[self._base_layer].is_valid(offset, length)

    @property
    def minimum_address(self) -> int:
        return 0

    @property
    def maximum_address(self) -> int:
        return self.config.get(
            "maximum_size", len(self._pages) * self._pdb_layer.page_size
        )

    @property
    def _pdb_layer(self) -> PdbMultiStreamFormat:
        if self._base_layer not in self._context.layers:
            raise PDBFormatException(
                self._base_layer,
                f"No PdbMultiStreamFormat layer found: {self._base_layer}",
            )
        result = self._context.layers[self._base_layer]
        if isinstance(result, PdbMultiStreamFormat):
            return result
        raise TypeError("Base layer is not PdbMultiStreamFormat")



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
import ntpath
from typing import List, Type, Optional, Iterator, Tuple

from volatility3.framework import constants, exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows.extensions import pe
from volatility3.plugins.windows import pslist, modules

vollog = logging.getLogger(__name__)


class PEDump(interfaces.plugins.PluginInterface):
    """Allows extracting PE Files from a specific address in a specific address space"""

    _required_framework_version = (2, 0, 0)

    # 2.0.0 - changed the signature of `dump_kernel_pe_at_base`
    _version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="modules", component=modules.Modules, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process IDs to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.IntRequirement(
                name="base",
                description="Base address to reconstruct a PE file",
                optional=False,
            ),
            requirements.BooleanRequirement(
                name="kernel_module",
                description="Extract from kernel address space.",
                default=False,
                optional=True,
            ),
        ]

    @classmethod
    def dump_pe(
        cls,
        context: interfaces.context.ContextInterface,
        pe_table_name: str,
        layer_name: str,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
        file_name: str,
        base: int,
    ) -> Optional[str]:
        """
        Returns the filename of the dump file or None
        """
        with open_method(file_name) as file_handle:
            try:
                dos_header = context.object(
                    pe_table_name + constants.BANG + "_IMAGE_DOS_HEADER",
                    offset=base,
                    layer_name=layer_name,
                )

                for offset, data in dos_header.reconstruct():
                    file_handle.seek(offset)
                    file_handle.write(data)
            except (
                OSError,
                exceptions.VolatilityException,
                OverflowError,
                ValueError,
            ) as excp:
                vollog.debug(f"Unable to dump PE file at offset {base}: {excp}")
                return None

            return file_handle.preferred_filename

    @classmethod
    def dump_ldr_entry(
        cls,
        context: interfaces.context.ContextInterface,
        pe_table_name: str,
        ldr_entry: interfaces.objects.ObjectInterface,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
        layer_name: Optional[str] = None,
        prefix: str = "",
    ) -> Optional[str]:
        """Extracts the PE file referenced an LDR_DATA_TABLE_ENTRY (DLL, kernel module) instance

        Args:
            context: the context to operate upon
            pe_table_name: the name for the symbol table containing the PE format symbols
            ldr_entry: the object representing the module
            open_method: class for constructing output files
            layer_name: the layer that the DLL lives within
            prefix: optional string to prepend to filename
        Returns:
            The output file name or None in the case of failure
        """
        try:
            name = ldr_entry.FullDllName.get_string()
        except exceptions.InvalidAddressException:
            name = "UnreadableDLLName"

        if layer_name is None:
            layer_name = ldr_entry.vol.layer_name

        file_name = f"{prefix}{ntpath.basename(name)}.{ldr_entry.vol.offset:#x}.{ldr_entry.DllBase:#x}.dmp"

        return cls.dump_pe(
            context,
            pe_table_name,
            layer_name,
            open_method,
            file_name,
            ldr_entry.DllBase,
        )

    @classmethod
    def dump_pe_at_base(
        cls,
        context: interfaces.context.ContextInterface,
        pe_table_name: str,
        layer_name: str,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
        proc_offset: int,
        pid: int,
        base: int,
    ) -> Optional[str]:
        file_name = f"PE.{proc_offset:#x}.{pid:d}.{base:#x}.dmp"

        return PEDump.dump_pe(
            context, pe_table_name, layer_name, open_method, file_name, base
        )

    @classmethod
    def dump_kernel_pe_at_base(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        pe_table_name: str,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
        base: int,
    ) -> Iterator[Tuple[int, str, str]]:
        """
        Extracts a PE file from kernel memory at the given base address
        """
        session_layers = modules.Modules.get_session_layers(
            context=context, kernel_module_name=kernel_module_name
        )

        session_layer_name = modules.Modules.find_session_layer(
            context, session_layers, base
        )

        if session_layer_name:
            system_pid = 4

            file_output = PEDump.dump_pe_at_base(
                context,
                pe_table_name,
                session_layer_name,
                open_method,
                0,
                system_pid,
                base,
            )

            if file_output:
                yield system_pid, "Kernel", file_output
        else:
            vollog.warning(
                "Unable to find a session layer with the provided base address mapped in the kernel."
            )

    @classmethod
    def dump_processes(
        cls, context, kernel, pe_table_name, open_method, filter_func, base
    ):
        """ """

        for proc in pslist.PsList.list_processes(
            context=context,
            kernel_module_name=kernel.name,
            filter_func=filter_func,
        ):
            pid = proc.UniqueProcessId
            proc_name = proc.ImageFileName.cast(
                "string",
                max_length=proc.ImageFileName.vol.count,
                errors="replace",
            )
            proc_layer_name = proc.add_process_layer()

            file_output = PEDump.dump_pe_at_base(
                context,
                pe_table_name,
                proc_layer_name,
                open_method,
                proc.vol.offset,
                pid,
                base,
            )

            if file_output:
                yield pid, proc_name, file_output

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]

        pe_table_name = intermed.IntermediateSymbolTable.create(
            self.context, self.config_path, "windows", "pe", class_types=pe.class_types
        )

        if self.config["kernel_module"] and self.config["pid"]:
            vollog.error("Only 'kernel-module' or 'pid' should be set, not both")
            return

        if not self.config["kernel_module"] and not self.config["pid"]:
            vollog.error("Either 'kernel-module' or 'pid' argument must be set")
            return

        if self.config["kernel_module"]:
            pe_files = self.dump_kernel_pe_at_base(
                context=self.context,
                kernel_module_name=self.config["kernel"],
                pe_table_name=pe_table_name,
                open_method=self.open,
                base=self.config["base"],
            )
        else:
            filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
            pe_files = self.dump_processes(
                self.context,
                kernel,
                pe_table_name,
                self.open,
                filter_func,
                self.config["base"],
            )

        for pid, proc_name, file_output in pe_files:
            yield (
                0,
                (
                    pid,
                    proc_name,
                    file_output,
                ),
            )

    def run(self):
        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("File output", str),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""Defines the automagic interfaces for populating the context before a plugin
runs.

Automagic objects attempt to automatically fill configuration values
that a user has not filled.
"""

import logging
from abc import ABCMeta
from typing import Any, List, Optional, Tuple, Type, Union

from volatility3.framework import constants, interfaces
from volatility3.framework.configuration import requirements

vollog = logging.getLogger(__name__)


class AutomagicInterface(
    interfaces.configuration.ConfigurableInterface, metaclass=ABCMeta
):
    """Class that defines an automagic component that can help fulfill
    `Requirements`

    These classes are callable with the following parameters:

    Args:
        context: The context in which to store configuration data that the automagic might populate
        config_path: Configuration path where the configurable's data under the context's config lives
        configurable: The top level configurable whose requirements may need satisfying
        progress_callback: An optional function accepting a percentage and optional description to indicate
            progress during long calculations

    .. note::

        The `context` provided here may be different to that provided during initialization.  The `context` provided at
        initialization should be used for local configuration of the automagic itself, the `context` provided during
        the call is to be populated by the automagic.
    """

    priority = 10
    """An ordering to indicate how soon this automagic should be run"""

    exclusion_list: List[str] = []
    """A list of plugin categories (typically operating systems) which the plugin will not operate on"""

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(context, config_path)
        for requirement in self.get_requirements():
            if not isinstance(
                requirement,
                (
                    interfaces.configuration.SimpleTypeRequirement,
                    requirements.ChoiceRequirement,
                    requirements.ListRequirement,
                    requirements.VersionRequirement,
                ),
            ):
                raise TypeError(
                    "Automagic requirements must be a SimpleTypeRequirement, ChoiceRequirement, ListRequirement or VersionRequirement"
                )

    def __call__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        requirement: interfaces.configuration.RequirementInterface,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[List[Any]]:
        """Runs the automagic over the configurable."""
        return []

    # TODO: requirement_type can be made UnionType[Type[T], Tuple[Type[T], ...]]
    #       once mypy properly supports Tuples in instance

    def find_requirements(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        requirement_root: interfaces.configuration.RequirementInterface,
        requirement_type: Union[
            Tuple[Type[interfaces.configuration.RequirementInterface], ...],
            Type[interfaces.configuration.RequirementInterface],
        ],
        shortcut: bool = True,
    ) -> List[Tuple[str, interfaces.configuration.RequirementInterface]]:
        """Determines if there is actually an unfulfilled `Requirement`
        waiting.

        This ensures we do not carry out an expensive search when there is no need for a particular `Requirement`

        Args:
            context: Context on which to operate
            config_path: Configuration path of the top-level requirement
            requirement_root: Top-level requirement whose subrequirements will all be searched
            requirement_type: Type of requirement to find
            shortcut: Only returns requirements that live under unsatisfied requirements

        Returns:
            A list of tuples containing the config_path, sub_config_path and requirement identifying the unsatisfied `Requirements`
        """
        sub_config_path = interfaces.configuration.path_join(
            config_path, requirement_root.name
        )
        results: List[Tuple[str, interfaces.configuration.RequirementInterface]] = []
        recurse = not shortcut
        if isinstance(requirement_root, requirement_type):
            if recurse or requirement_root.unsatisfied(context, config_path):
                results.append((sub_config_path, requirement_root))
        else:
            recurse = True
        if recurse:
            for subreq in requirement_root.requirements.values():
                results += self.find_requirements(
                    context, sub_config_path, subreq, requirement_type, shortcut
                )
        return results


class StackerLayerInterface(metaclass=ABCMeta):
    """Class that takes a lower layer and attempts to build on it.

    stack_order determines the order (from low to high) that stacking
    layers should be attempted lower levels should have lower
    `stack_orders`
    """

    stack_order = 0
    """The order in which to attempt stacking, the lower the earlier"""
    exclusion_list: List[str] = []
    """The list operating systems/first-level plugin hierarchy that should exclude this stacker"""

    @classmethod
    def stack(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[interfaces.layers.DataLayerInterface]:
        """Method to determine whether this builder can operate on the named
        layer.  If so, modify the context appropriately.

        Returns the name of any new layer stacked on top of this layer or None.  The stacking is therefore strictly
        linear rather than tree driven.

        Configuration options provided by the context are ignored, and defaults are to be used by this method
        to build a space where possible.

        Args:
           context: Context in which to construct the higher layer
           layer_name: Name of the layer to stack on top of
           progress_callback: A callback function to indicate progress through a scan (if one is necessary)
        """

    @classmethod
    def stacker_slow_warning(cls):
        vollog.warning(
            "Reads to this layer are slow, it's recommended to use the layerwriter plugin once to produce a raw file"
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from typing import Any, List, Optional, Tuple, Union
from enum import Enum

from volatility3.cli.volshell import generic
from volatility3.framework import constants, interfaces
from volatility3.framework.configuration import requirements
from volatility3.plugins.linux import pslist


# Could import the enum from psscan.py to avoid code duplication
class DescExitStateEnum(Enum):
    """Enum for linux task exit_state as defined in include/linux/sched.h"""

    TASK_RUNNING = 0x00000000
    EXIT_DEAD = 0x00000010
    EXIT_ZOMBIE = 0x00000020
    EXIT_TRACE = EXIT_ZOMBIE | EXIT_DEAD


class Volshell(generic.Volshell):
    """Shell environment to directly interact with a linux memory image."""

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel", description="Linux kernel module"
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.IntRequirement(
                name="pid", description="Process ID", optional=True
            ),
            requirements.VersionRequirement(
                name="generic_volshell",
                component=generic.Volshell,
                version=(1, 0, 0),
            ),
        ] + super().get_requirements()

    def change_task(self, pid=None):
        """Change the current process and layer, based on a process ID"""
        tasks = self.list_tasks()
        for task in tasks:
            if task.pid == pid:
                process_layer = task.add_process_layer()
                if process_layer is not None:
                    self.change_layer(process_layer)
                    return None
                print(f"Layer for task ID {pid} could not be constructed")
                return None
        print(f"No task with task ID {pid} found")

    def get_process(self, pid=None, virtaddr=None, physaddr=None):
        """Return the task_struct object that matches the pid. If a physical or a virtual address is provided, construct the task_struct object at said address. Only one parameter is allowed.

        Args:
            pid (int, optional): PID to search for
            virtaddr (int, optional): Virtual address to construct object at
            physaddr (int, optional): Physical address to construct object at

        Returns:
            ObjectInterface: task_struct Object
        """

        if sum(1 if x is not None else 0 for x in [pid, virtaddr, physaddr]) != 1:
            print("Only one parameter is accepted")
            return None

        vmlinux_module_name = self.config["kernel"]
        vmlinux = self.context.modules[vmlinux_module_name]

        kernel_layer_name = vmlinux.layer_name
        kernel_layer = self.context.layers[kernel_layer_name]

        memory_layer_name = kernel_layer.dependencies[0]

        task_struct_symbol = vmlinux.symbol_table_name + constants.BANG + "task_struct"

        if virtaddr is not None:
            task = self.context.object(
                task_struct_symbol,
                layer_name=kernel_layer_name,
                offset=virtaddr,
            )

        if physaddr is not None:
            task = self.context.object(
                task_struct_symbol,
                layer_name=memory_layer_name,
                offset=physaddr,
                native_layer_name=kernel_layer_name,
            )

        if physaddr is not None or virtaddr is not None:
            try:
                DescExitStateEnum(task.exit_state)
            except ValueError:
                print(
                    f"task_struct @ {hex(task.vol.offset)} as exit_state {task.exit_state} is likely not valid"
                )

            if not (0 < task.pid < 65535):
                print(
                    f"task_struct @ {hex(task.vol.offset)} as pid {task.pid} is likely not valid"
                )

            return task

        if pid is not None:
            tasks = self.list_tasks()
            for task in tasks:
                if task.pid == pid:
                    return task
            print(f"No task with task ID {pid} found")

        return None

    def list_tasks(self):
        """Returns a list of task objects from the primary layer"""
        # We always use the main kernel memory and associated symbols
        return list(pslist.PsList.list_tasks(self.context, self.current_kernel_name))

    def construct_locals(self) -> List[Tuple[List[str], Any]]:
        result = super().construct_locals()
        result += [
            (["ct", "change_task", "cp"], self.change_task),
            (["lt", "list_tasks", "ps"], self.list_tasks),
            (["gp", "get_process", "get_task"], self.get_process),
            (["symbols"], self.context.symbol_space[self.current_symbol_table]),
        ]
        if self.config.get("pid", None) is not None:
            self.change_task(self.config["pid"])
        return result

    def display_type(
        self,
        object: Union[
            str, interfaces.objects.ObjectInterface, interfaces.objects.Template
        ],
        offset: Optional[int] = None,
    ):
        """Display Type describes the members of a particular object in alphabetical order"""
        if isinstance(object, str):
            if constants.BANG not in object:
                object = self.current_symbol_table + constants.BANG + object
        return super().display_type(object, offset)

    def display_symbols(self, symbol_table: Optional[str] = None):
        """Prints an alphabetical list of symbols for a symbol table"""
        if symbol_table is None:
            symbol_table = self.current_symbol_table
        return super().display_symbols(symbol_table)

    @property
    def current_layer(self):
        if self.__current_layer is None:
            self.__current_layer = self.kernel.layer_name
        return self.__current_layer



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from volatility3.framework.configuration import requirements as requirements



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

# This file contains the Deprecation class used to deprecate methods in an orderly manner

import warnings
import functools
import inspect

from typing import Callable, Tuple

from volatility3.framework import interfaces, exceptions, versionutils


def method_being_removed(message: str, removal_date: str):
    """A decorator for marking functions as being removed in the future and without a replacement.
       Callers to this function should explicitly list the API paths that should be used instead.

    Args:
        message: A message added to the standard deprecation warning. Should include the replacement API paths
        removal_date: A YYYY-MM-DD formatted date of when the function will be removed from the framework
    """

    def decorator(deprecated_func):
        @functools.wraps(deprecated_func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"This API ({deprecated_func.__module__}.{deprecated_func.__qualname__}) will be removed in the first release after {removal_date}. {message}",
                FutureWarning,
            )
            return deprecated_func(*args, **kwargs)

        return wrapper

    return decorator


def deprecated_method(
    replacement: Callable,
    removal_date: str,
    replacement_version: Tuple[int, int, int] = None,
    additional_information: str = "",
):
    """A decorator for marking functions as deprecated.

    Args:
        replacement: The replacement function overriding the deprecated API, in the form of a Callable (typically a method)
        removal_date: A YYYY-MM-DD formatted date of when the function will be removed from the framework
        replacement_version: The "replacement" base class version that the deprecated method expects before proxying to it. This implies that "replacement" is a method from a class that inherits from VersionableInterface.
        additional_information: Information appended at the end of the deprecation message
    """

    def decorator(deprecated_func):
        @functools.wraps(deprecated_func)
        def wrapper(*args, **kwargs):
            nonlocal replacement, replacement_version, additional_information
            # Prevent version mismatches between deprecated (proxy) methods and the ones they proxy
            if (
                replacement_version is not None
                and callable(replacement)
                and hasattr(replacement, "__self__")
            ):
                replacement_base_class = replacement.__self__

                # Verify that the base class inherits from VersionableInterface
                if inspect.isclass(replacement_base_class) and issubclass(
                    replacement_base_class,
                    interfaces.configuration.VersionableInterface,
                ):
                    # SemVer check
                    if not versionutils.matches_required(
                        replacement_version, replacement_base_class.version
                    ):
                        raise exceptions.VersionMismatchException(
                            deprecated_func,
                            replacement_base_class,
                            replacement_version,
                            "This is a bug, the deprecated call needs to be removed and the caller needs to update their code to use the new method.",
                        )

            deprecation_msg = f'Method "{deprecated_func.__module__ + "." + deprecated_func.__qualname__}" is deprecated and will be removed in the first release after {removal_date}, use "{replacement.__module__ + "." + replacement.__qualname__}" instead. {additional_information}'
            warnings.warn(deprecation_msg, FutureWarning)
            # Return the wrapped function with its original arguments
            return deprecated_func(*args, **kwargs)

        return wrapper

    return decorator


def renamed_class(deprecated_class_name: str, message: str, removal_date: str):
    """A decorator for marking classes as being renamed and removed in the future.
       Callers to this function should explicitly update to use the other plugins instead.

    Args:
        deprecated_class_name: The name of the class being deprecated
        message: A message added to the standard deprecation warning. Should include the replacement API paths
        removal_date: A YYYY-MM-DD formatted date of when the function will be removed from the framework
    """

    def decorator(replacement_func):
        @functools.wraps(replacement_func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"This plugin ({deprecated_class_name}) has been renamed and will be removed in the first release after {removal_date}. {message}",
                FutureWarning,
            )
            return replacement_func(*args, **kwargs)

        return wrapper

    return decorator


class PluginRenameClass:
    """Class to move all classmethod invocations (for when a plugin has been moved)"""

    def __init_subclass__(cls, replacement_class, removal_date, **kwargs):
        deprecated_class_name = f"{cls.__module__}.{cls.__qualname__}"
        super().__init_subclass__(**kwargs)
        for attr, value in replacement_class.__dict__.items():
            if isinstance(value, classmethod) and attr != "get_requirements":
                setattr(
                    cls,
                    attr,
                    classmethod(
                        renamed_class(
                            deprecated_class_name=deprecated_class_name,
                            removal_date=removal_date,
                            message=f"Please ensure all method calls to this plugin are replaced with calls to {replacement_class.__module__}.{replacement_class.__qualname__}",
                        )(value.__func__)
                    ),
                )
            else:
                if attr == "run":
                    setattr(
                        cls,
                        attr,
                        method_being_removed(
                            removal_date=removal_date,
                            message=f"This plugin has been renamed, please call {replacement_class.__module__}.{replacement_class.__qualname__} rather than {deprecated_class_name}.",
                        )(value),
                    )
                elif not attr.startswith("__"):
                    setattr(cls, attr, value)
        return super(PluginRenameClass).__init_subclass__(**kwargs)



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
import threading
from typing import Any, Dict, IO, List, Optional, Union

from volatility3.framework import constants, exceptions, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import resources

vollog = logging.getLogger(__name__)


class BufferDataLayer(interfaces.layers.DataLayerInterface):
    """A DataLayer class backed by a buffer in memory, designed for testing and
    swift data access."""

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        name: str,
        buffer: bytes,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(
            context=context, config_path=config_path, name=name, metadata=metadata
        )
        self._buffer = buffer

    @property
    def maximum_address(self) -> int:
        """Returns the largest available address in the space."""
        return len(self._buffer) - 1

    @property
    def minimum_address(self) -> int:
        """Returns the smallest available address in the space."""
        return 0

    def is_valid(self, offset: int, length: int = 1) -> bool:
        """Returns whether the offset is valid or not."""
        return bool(
            self.minimum_address <= offset <= self.maximum_address
            and self.minimum_address <= offset + length - 1 <= self.maximum_address
        )

    def read(self, address: int, length: int, pad: bool = False) -> bytes:
        """Reads the data from the buffer."""
        if not self.is_valid(address, length):
            invalid_address = address
            if self.minimum_address < address <= self.maximum_address:
                invalid_address = self.maximum_address + 1
            raise exceptions.InvalidAddressException(
                self.name, invalid_address, "Offset outside of the buffer boundaries"
            )
        return self._buffer[address : address + length]

    def write(self, address: int, data: bytes):
        """Writes the data from to the buffer."""
        self._buffer = (
            self._buffer[:address] + data + self._buffer[address + len(data) :]
        )

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # No real requirements (only the buffer).  Need to figure out if there's a better way of representing this
        return [
            requirements.BytesRequirement(
                name="buffer",
                description="The direct bytes to interact with",
                optional=False,
            )
        ]


class DummyLock:
    def __enter__(self) -> None:
        pass

    def __exit__(self, type, value, traceback) -> None:
        pass


class FileLayer(interfaces.layers.DataLayerInterface):
    """a DataLayer backed by a file on the filesystem."""

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(
            context=context, config_path=config_path, name=name, metadata=metadata
        )

        self._write_warning = False
        self._location = self.config["location"]
        self._accessor = resources.ResourceAccessor()
        self._file_: Optional[IO[Any]] = None
        self._size: Optional[int] = None
        self._maximum_address: Optional[int] = None
        # Construct the lock now (shared if made before threading) in case we ever need it
        self._lock: Union[DummyLock, threading.Lock] = DummyLock()
        if constants.PARALLELISM == constants.Parallelism.Threading:
            self._lock = threading.Lock()
        # Instantiate the file to throw exceptions if the file doesn't open
        _ = self._file

    @property
    def location(self) -> str:
        """Returns the location on which this Layer abstracts."""
        return self._location

    @property
    def _file(self) -> IO[Any]:
        """Property to prevent the initializer storing an unserializable open
        file (for context cloning)"""
        # FIXME: Add "+" to the mode once we've determined whether write mode is enabled
        mode = "rb"
        self._file_ = self._file_ or self._accessor.open(self._location, mode)
        return self._file_

    @property
    def maximum_address(self) -> int:
        """Returns the largest available address in the space."""
        # Zero based, so we return the size of the file minus 1
        if self._maximum_address:
            return self._maximum_address
        with self._lock:
            orig = self._file.tell()
            self._file.seek(0, 2)
            self._size = self._file.tell()
            self._file.seek(orig)
            self._maximum_address = self._size - 1
        return self._maximum_address

    @property
    def minimum_address(self) -> int:
        """Returns the smallest available address in the space."""
        return 0

    def is_valid(self, offset: int, length: int = 1) -> bool:
        """Returns whether the offset is valid or not."""
        if length <= 0:
            raise ValueError("Length must be positive")
        return bool(
            self.minimum_address <= offset <= self.maximum_address
            and self.minimum_address <= offset + length - 1 <= self.maximum_address
        )

    def read(self, offset: int, length: int, pad: bool = False) -> bytes:
        """Reads from the file at offset for length."""
        if not self.is_valid(offset, length):
            invalid_address = offset
            if self.minimum_address < offset <= self.maximum_address:
                invalid_address = self.maximum_address + 1
            raise exceptions.InvalidAddressException(
                self.name, invalid_address, "Offset outside of the buffer boundaries"
            )

        # TODO: implement locking for multi-threading
        with self._lock:
            self._file.seek(offset)
            data = self._file.read(length)

        if len(data) < length:
            if pad:
                data += b"\x00" * (length - len(data))
            else:
                raise exceptions.InvalidAddressException(
                    self.name,
                    offset + len(data),
                    "Could not read sufficient bytes from the " + self.name + " file",
                )
        return data

    def write(self, offset: int, data: bytes) -> None:
        """Writes to the file.

        This will technically allow writes beyond the extent of the file
        """
        if not self._file.writable():
            if not self._write_warning:
                self._write_warning = True
                vollog.warning(f"Try to write to unwritable layer: {self.name}")
            return None
        if not self.is_valid(offset, len(data)):
            invalid_address = offset
            if self.minimum_address < offset <= self.maximum_address:
                invalid_address = self.maximum_address + 1
            raise exceptions.InvalidAddressException(
                self.name,
                invalid_address,
                "Data segment outside of the " + self.name + " file boundaries",
            )
        with self._lock:
            self._file.seek(offset)
            self._file.write(data)

    def __getstate__(self) -> Dict[str, Any]:
        """Do not store the open _file_ attribute, our property will ensure the
        file is open when needed.

        This is necessary for multi-processing
        """
        self._file_ = None
        return self.__dict__

    def destroy(self) -> None:
        """Closes the file handle."""
        self._file.close()

    def __exit__(self, type, value, traceback) -> None:
        self.destroy()

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [requirements.StringRequirement(name="location", optional=False)]
